# import matplotlib
# matplotlib.use('Qt5Agg')  # Prevents `Invalid DISPLAY variable` errors

import pytest
import tempfile

from calliope import Model
from calliope.utils import AttrDict

from calliope import analysis

from . import common
from .common import assert_almost_equal, solver, solver_io

import matplotlib.pyplot as plt
plt.switch_backend('agg')  # Prevents `Invalid DISPLAY variable` errors


class TestModel:
    @pytest.fixture(scope='module')
    def model(self):
        locations = """
            locations:
                1:
                    techs: ['ccgt', 'demand_power']
                    override:
                        ccgt:
                            constraints:
                                e_cap.max: 100
                        demand_power:
                            constraints:
                                r: -50
            metadata:
                map_boundary: [-10, 35, 5, 45]
                location_coordinates:
                    1: [40, -2]
            links:
        """
        config_run = """
            mode: plan
            model: ['{techs}', '{locations}']
            subset_t: ['2005-01-01', '2005-01-02']
        """
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(locations.encode('utf-8'))
            f.read()
            override_dict = AttrDict({
                'solver': solver,
                'solver_io': solver_io,
            })
            model = common.simple_model(config_run=config_run,
                                        config_locations=f.name,
                                        override=override_dict)
        model.run()
        return model

    @pytest.fixture(scope='module')
    def builtin_model(self):
        model = Model()
        model.run()
        return model

    def test_plot_carrier_production(self, model):
        # Just make sure this doesn't raise any exceptions
        analysis.plot_carrier_production(model.solution)

    def test_plot_timeseries(self, model):
        # Just make sure this doesn't raise any exceptions
        analysis.plot_timeseries(model.solution,
                                 model.solution['e'].loc[dict(c='power')].sum(dim='x'),
                                 carrier='power', demand='demand_power')

    def test_plot_installed_capacities(self, model):
        # Just make sure this doesn't raise any exceptions
        analysis.plot_installed_capacities(model.solution)

    def test_plot_transmission(self, model):
        # Just make sure this doesn't raise any exceptions
        analysis.plot_transmission(model.solution, map_resolution='c')

    def test_get_delivered_cost(self, model):
        # TODO this should be tested with a more complex model
        assert_almost_equal(analysis.get_delivered_cost(model.solution), 0.1)

    def test_get_levelized_cost(self, model):
        lcoe = analysis.get_levelized_cost(model.solution)
        assert_almost_equal(lcoe.at['ccgt'], 0.1)

    def test_get_group_share(self, model):
        # TODO this should be tested with a more complex model
        share = analysis.get_group_share(model.solution, techs=['ccgt'])
        assert share == 1.0

    def test_get_unmet_demand_hours(self, builtin_model):
        # TODO this should be tested with a more complex model
        unmet = analysis.get_unmet_demand_hours(builtin_model.solution)
        assert unmet == 1

    def test_recompute_levelized_costs(self, model):
        # Cost in solution
        sol = model.solution
        assert_almost_equal(sol['summary'].to_pandas().loc['ccgt', 'levelized_cost_monetary'], 0.1)
        # Recomputed cost must be the same
        dm = analysis.SolutionModel(model.solution)
        recomputed = dm.recompute_levelized_costs('ccgt')
        assert_almost_equal(recomputed['total'], 0.1)

    def test_recompute_levelized_costs_after_changes(self, model):
        # Make changes
        dm = analysis.SolutionModel(model.solution)
        dm.config_model.techs.ccgt.costs.monetary.e_cap = 50
        dm.config_model.techs.ccgt.costs.monetary.om_fuel = 1.0
        # Recomputed cost
        recomputed = dm.recompute_levelized_costs('ccgt')
        assert_almost_equal(recomputed['total'], 1.0, tolerance=0.001)



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

time_funcs.py
~~~~~~~~~~~~~

Functions to process time series data.

"""

import logging

import pandas as pd
import xarray as xr
from xarray.ufuncs import fabs  # pylint: disable=no-name-in-module

from . import utils
from . import time_clustering


def normalized_copy(data):
    """
    Return a copy of data, with the absolute taken and normalized to 0-1.

    The maximum across all regions and timesteps is used to normalize.

    """
    ds = data.copy(deep=True)  # Work off a copy
    data_vars_in_t = [v for v in time_clustering._get_datavars(data)
                      if 't' in data[v].dims]
    for var in data_vars_in_t:
        for y in ds.coords['y'].values:
            # Get max across all regions to normalize against
            norm_max = fabs(ds[var].loc[{'y': y}]).max()
            for x in ds.coords['x'].values:
                df = ds[var].loc[{'x': x, 'y': y}]
                ds[var].loc[{'x': x, 'y': y}] = fabs(df) / norm_max
    return ds


def _copy_non_t_vars(data0, data1):
    """Copies non-t-indexed variables from data0 into data1, then
    returns data1"""
    non_t_vars = [v for v in data0.data_vars
                  if 't' not in data0[v].dims]
    # Manually copy over variables not in `t`. If we don't do this,
    # these vars get polluted with a superfluous `t` dimension
    for v in non_t_vars:
        data1[v] = data0[v]
    return data1


def _combine_datasets(data0, data1):
    """Concatenates data0 and data1 along the t dimension"""
    data_new = xr.concat([data0, data1], dim='t')
    # Ensure time dimension is ordered
    data_new = data_new.loc[{'t': data_new.t.to_pandas().index.sort_values()}]

    return data_new


def apply_clustering(data, timesteps, clustering_func, how, normalize=True, **kwargs):
    """
    Apply the given clustering function to the given data.

    Parameters
    ----------
    data : xarray.Dataset
    timesteps : pandas.DatetimeIndex or list of timesteps or None
    clustering_func : str
        Name of clustering function.
    how : str
        How to map clusters to data. 'mean' or 'closest'.
    normalize : bool, optional
        If True (default), data is normalized before clustering is applied,
        using :func:`~calliope.time_funcs.normalized_copy`.
    **kwargs : optional
        Arguments passed to clustering_func.

    Returns
    -------
    data_new_scaled : xarray.Dataset

    """
    # Only apply clustering function on subset of masked timesteps
    if timesteps is None:
        data_to_cluster = data
    else:
        data_to_cluster = data.loc[{'t': timesteps}]

    if normalize:
        data_normalized = normalized_copy(data_to_cluster)
    else:
        data_normalized = data_to_cluster

    # Get function from `clustering_func` string
    func = utils.plugin_load(clustering_func, builtin_module='time_clustering')

    result = func(data_normalized, **kwargs)
    clusters = result[0]  # Ignore other stuff returned

    data_new = time_clustering.map_clusters_to_data(data_to_cluster, clusters,
                                                    how=how)

    if timesteps is None:
        data_new = _copy_non_t_vars(data, data_new)
    else:
        # Drop timesteps from old data
        data_new = _copy_non_t_vars(data, data_new)
        data_new = _combine_datasets(data.drop(timesteps, dim='t'), data_new)
        data_new = _copy_non_t_vars(data, data_new)

    # Scale the new/combined data so that the mean for each (x, y, variable)
    # combination matches that from the original data
    data_new_scaled = data_new.copy(deep=True)
    data_vars_in_t = [v for v in time_clustering._get_datavars(data)
                      if 't' in data[v].dims]
    for var in data_vars_in_t:
        scale_to_match_mean = (data[var].mean(dim='t') / data_new[var].mean(dim='t')).fillna(0)
        data_new_scaled[var] = data_new[var] * scale_to_match_mean

    return data_new_scaled


_RESAMPLE_METHODS = {
    '_weights': 'mean',
    '_time_res': 'sum',
    'r': 'sum',
    'e_eff': 'mean',
}


def resample(data, timesteps, resolution):
    data_new = data.copy(deep=True)
    if timesteps is not None:
        data_new = data_new.loc[{'t': timesteps}]

    # First create a new resampled dataset of the correct size by
    # using first-resample, which should be a quick way to achieve this
    data_rs = data_new.resample(resolution, dim='t', how='first')

    timestep_vars = [v for v in data_new.data_vars
                     if 't' in data_new[v].dims]

    # Resampling adds spurious `t` dimension to non-t vars, correct that
    for v in data_rs.data_vars:
        if v not in timestep_vars:
            data_rs[v] = data[v]

    for var in timestep_vars:
        if var in _RESAMPLE_METHODS:
            how = _RESAMPLE_METHODS[var]
            data_rs[var] = data_new[var].resample(resolution, dim='t', how=how)
        else:
            # If we don't know how to resample a var, we drop it
            logging.error('Dropping {} because it has no resampling method.'.format(var))
            data_rs = data_rs.drop(var)

    # Get rid of the filled-in NaN timestamps
    data_rs = data_rs.dropna(dim='t', how='all')
    data_rs.attrs['opmode_safe'] = True  # Resampling still permits operational mode

    if timesteps is not None:
        # Combine leftover parts of passed in data with new data
        data_rs = _copy_non_t_vars(data, data_rs)
        data_rs = _combine_datasets(data.drop(timesteps, dim='t'), data_rs)
        data_rs = _copy_non_t_vars(data, data_rs)
        # Having timesteps with different lengths does not permit operational mode
        data_rs.attrs['opmode_safe'] = False

    return data_rs


def drop(data, timesteps, padding=None):
    """
    Drop timesteps from data, with optional padding
    around into the contiguous areas encompassed by the timesteps.

    """
    if padding:
        ts_per_day = time_clustering._get_timesteps_per_day(data)
        freq = '{}H'.format(24 / ts_per_day)

        # Series of 1 where timesteps 'exist' and 0 where they don't
        s = (pd.Series(1, index=timesteps)
               .reindex(pd.date_range(timesteps[0], timesteps[-1], freq=freq))
               .fillna(0))

        # Blocks of contiguous 1's in the series
        blocks = (s != s.shift()).cumsum().drop(s[s==0].index)

        # Groups of contiguous areas
        groups = blocks.groupby(blocks).apply(lambda x: (x.index[0], x.index[-1]))

        # Reduce size of each block by `padding` on both sides
        padding = pd.Timedelta(padding)
        dt_indices = [pd.date_range(g[0] + padding, g[1] - padding, freq=freq)
                      for g in groups]

        # Concatenate the DatetimeIndexes by using dummy Series
        timesteps = pd.concat([pd.Series(0, index=i) for i in dt_indices]).index

    # 'Distribute weight' of the dropped timesteps onto the remaining ones
    dropped_weight = data._weights.loc[{'t': timesteps}].sum()

    data = data.drop(timesteps, dim='t')

    data['_weights'] = data['_weights'] + (dropped_weight / len(data['_weights']))

    return data



import pytest  # pylint: disable=unused-import

# import pandas as pd

# import calliope


# class TestMaskWhereZero:
#     @pytest.fixture(scope='module')
#     def testdata(self):
#         d = calliope.utils.AttrDict()
#         df1 = pd.DataFrame({'a': list(range(1, 11)),
#                             'b': list(range(20, 30))})
#         df2 = pd.DataFrame({'a': [0, 0, 0, 0, 1, 2, 3, 4, 0, 0],
#                             'b': [0, 0, 0, 1, 2, 1, 1, 2, 0, 0]})
#         df3 = pd.DataFrame({'a': [0, 0, 0, 0, 1, 2, 3, 4, 1, 0],
#                             'b': [0, 0, 0, 1, 2, 1, 1, 2, 0, 0]})
#         df4 = pd.DataFrame({'a': [0, 0, 0, 0, 1, 2, 3, 4, 1, 2],
#                             'b': [0, 0, 0, 1, 2, 1, 1, 2, 0, 0]})
#         df5 = pd.DataFrame({'a': [1, 2, 2, 3, 1, 2, 0, 0, 0, 0],
#                             'b': [0, 0, 0, 1, 2, 1, 1, 0, 0, 0]})
#         d.set_key('r.tech1', df1)
#         d.set_key('r.tech2', df2)
#         d.set_key('r.tech3', df3)
#         d.set_key('r.tech4', df4)
#         d.set_key('r.tech5', df5)
#         d.set_key('e_eff.tech1', df1 / 100.0)
#         return d

#     def test_mask_zero_no_zeros(self, testdata):
#         data = testdata
#         mask = calliope.time_masks.mask_zero(data, 'tech1')
#         assert mask.tolist() == [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

#     def test_mask_zero_start_and_end_zero(self, testdata):
#         data = testdata
#         mask = calliope.time_masks.mask_zero(data, 'tech2')
#         assert mask.tolist() == [1, 1, 1, 0, 0, 0, 0, 0, 1, 1]
#         res = calliope.time_tools.masks_to_resolution_series([mask])
#         assert res.tolist() == [3, -1, -1, 0, 0, 0, 0, 0, 2, -1]

#     def test_mask_zero_one_zero(self, testdata):
#         data = testdata
#         mask = calliope.time_masks.mask_zero(data, 'tech3')
#         assert mask.tolist() == [1, 1, 1, 0, 0, 0, 0, 0, 0, 1]
#         res = calliope.time_tools.masks_to_resolution_series([mask])
#         assert res.tolist() == [3, -1, -1, 0, 0, 0, 0, 0, 0, 0]

#     def test_mask_zero_start_zero(self, testdata):
#         data = testdata
#         mask = calliope.time_masks.mask_zero(data, 'tech4')
#         assert mask.tolist() == [1, 1, 1, 0, 0, 0, 0, 0, 0, 0]
#         res = calliope.time_tools.masks_to_resolution_series([mask])
#         assert res.tolist() == [3, -1, -1, 0, 0, 0, 0, 0, 0, 0]

#     def test_mask_zero_end_zero(self, testdata):
#         data = testdata
#         mask = calliope.time_masks.mask_zero(data, 'tech5')
#         assert mask.tolist() == [0, 0, 0, 0, 0, 0, 0, 1, 1, 1]
#         res = calliope.time_tools.masks_to_resolution_series([mask])
#         assert res.tolist() == [0, 0, 0, 0, 0, 0, 0, 3, -1, -1]


# class TestDownsampling:
#     def test_infinity_test_no_inf(self):
#         pass

#     def test_infinity_test_all_inf(self):
#         pass

#     def test_infinity_test_mixed_inf(self):
#         pass

#     def test_reduce_average(self, df):
#         pass

#     def test_reduce_sum(self, df):
#         pass

#     def test_reduce_cut(self, df):
#         pass



"""
Improved networkx plotting functionality

Adapted from:

Source: https://networkx.lanl.gov/trac/ticket/434
Author: Aric Hagberg, https://github.com/hagberg
License: BSD license, see LICENSE_NetworkX

"""


import networkx as nx
import matplotlib.patches as patches
import matplotlib.collections as collections
import numpy as np
import math
import matplotlib.cbook as cb
import matplotlib.colors as mcolors
import matplotlib as mpl
import matplotlib.cm as cm


def get_color_dict(color,item_list,vmin=None,vmax=None,cmap=None):
    """ Determine color rgb color values given a data list.

    This function is used to take a sequence of data and convert it
    into a dictionary which of possible rgb values. It can take a
    number of different types of data explained below. If no colormap
    is given it will return the grayscale for scalar values.

    Parameters:
    -----------
    color: A string, scalar, or iterable of either.
           This can be a color in a variety of formats. A matplotlib
           a single color, which can be a matplotlib color specification
           e.g. 'rgbcmykw', a hex color string '#00FFFF', a standard
           html name like 'aqua', or a numeric value to be scaled with
           vmax or vmin. It can also be a list of any of these types
           as well as a dictionary of any of these types.
    item_list: a list
               A list of keys which correspond to the values given in
               color.
    vmin: A scalar
          The minimum value if scalar values are given for color
    vmax: A scalar
          The maximum value if scalar values are given for color
    cmap: A matplotlib colormap
          A colormap to be used if scalar values are given.

    Returns:
    -------
    color_dict: dict
                A dictionary of rgb colors keyed by values in item_list
    """

    CC = mcolors.ColorConverter()

    if cb.is_string_like(color):
        return {}.fromkeys(item_list,CC.to_rgb(color))
    elif cb.is_scalar(color):
        CN = mcolors.Normalize(0.0,1.0)
        if cmap is not None:
            return {}.fromkeys(item_list,cmap(CN(color)))
        else:
            return {}.fromkeys(item_list,CC.to_rgb(str(CN(color))))
    elif cb.iterable(color) and not cb.is_string_like(color):
        try:
            vals = [color[i] for i in item_list]
        except:
            vals = color
        keys = item_list
        if len(item_list)>len(vals):
            raise nx.NetworkXError("Must provide a value for each item")
        if cb.alltrue([cb.is_string_like(c) for c in vals]):
            return dict(list(zip(keys,[CC.to_rgb(v) for v in vals])))
        elif cb.alltrue([cb.is_scalar(c) for c in vals]):
            if vmin is None:
                vmin = float(min(vals))
            if vmax is None:
                vmax = float(max(vals))
            CN = mcolors.Normalize(vmin,vmax)
            if cmap is not None:
                return dict(list(zip(keys,[cmap(CN(v)) for v in vals])))
            else:
                return dict(list(zip(keys,[CC.to_rgb(str(CN(v))) for v in vals])))
        elif cb.alltrue([cb.iterable(c) and not cb.is_string(c) for c in vals]):
            return dict(list(zip(keys,vals)))
    else:
        raise nx.NetworkXError("Could not convert colors")


def is_weighted(G):
    """ Determine if a graph G is weighted

    Checks each edge to see if it has attribute 'weight' if it does
    return True, otherwise false. This checks if the entire graph is
    weighted not partial.

    Parameters:
    ----------
    G: A networkx Graph

    Returns:
    --------
    weighted :  A bool
       Determines whether the graph is weighted.
    """
    weighted = True
    for (u,v) in G.edges():
        weighted = weighted and ('weight' in G.edge[u][v])
        if not weighted:
            return weighted
    return weighted


def edge_width_weight(G,edgelist=None):
    """Automatically calculate a normalized reasonable line width for
    a weighted graph

    Parameters:
    -----------
    G: A networkx Graph
    edgelist: A list
      Edges to calculate the weights for if None, usesall edges
    Returns:
    --------
    weight_dict: A dictionary
       Line weights that displays nicely in matplotlib.
    """
    if edgelist is None:
        edgelist = G.edges()
    lw = {}
    for (u,v) in edgelist:
        lw[(u,v)] = G.edge[u][v]['weight']
    maxw = max(lw.values())
    minw = float(min(lw.values())) #to ensure floats later
    return dict(list(zip(list(lw.keys()), \
                    list(map((lambda x: 19.5*((x-minw)/(maxw-minw)) + 0.5), \
                        list(lw.values()))))))


def edge_color_weight(G,edgelist=None):
    """Automatically calculate a normalized reasonable color for
    a weighted graph

    Parameters:
    -----------
    G: A networkx Graph
    edgelist: A list
      Edges to calculate the weights for if None, uses all edges
    Returns:
    --------
    weight_dict: A dictionary
       Values between 0-1 that displays nicely in matplotlib.
    """
    cl = {}
    if edgelist is None:
        edgelist=G.edges()
    for (u,v) in edgelist:
        cl[(u,v)] = G[u][v]['weight']
    maxw = max(cl.values())
    minw = float(min(cl.values()))
    return dict(list(zip(list(cl.keys()), \
                    list(map((lambda x: (x-minw)/(maxw - minw)),list(cl.values()))))))


def draw_networkx_nodes(G, pos,
                        nodelist=None,
                        node_size=300,
                        node_color='r',
                        node_shape='o',
                        alpha=1.0,
                        cmap=None,
                        vmin=None,
                        vmax=None,
                        ax=None,
                        linewidth=None,
                        zorder=None,
                        **kwds):
    """Draw the nodes of the graph G.

    This draws only the nodes of the graph G.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       If not specified a spring layout positioning will be computed.
       See networkx.layout for functions that compute node positions.

    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.

    nodelist: list, optional
       Draw only specified nodes (default G.nodes())

    edgelist: list
       Draw only specified edges(default=G.edges())

    node_size: scalar or array
       Size of nodes (default=300).  If an array is specified it must be the
       same length as nodelist.

    node_color: color string, or array of floats
       Node color. Can be a single color format string (default='r'),
       or a  sequence of colors with the same length as nodelist.
       If numeric values are specified they will be mapped to
       colors using the cmap and vmin,vmax parameters.  See
       matplotlib.scatter for more details.

    node_shape:  string
       The shape of the node.  Specification is as matplotlib.scatter
       marker, one of 'so^>v<dph8' (default='o').

    alpha: float
       The node transparency (default=1.0)

    cmap: Matplotlib colormap
       Colormap for mapping intensities of nodes (default=None)

    vmin,vmax: floats
       Minimum and maximum for node colormap scaling (default=None)

    width`: float
       Line width of edges (default =1.0)


    Notes
    -----
    Any keywords not listed above are passed through to Matplotlib's
    scatter function.

    Examples
    --------
    >>> G=nx.dodecahedral_graph()
    >>> nodes=nx.draw_networkx_nodes(G,pos=nx.spring_layout(G))

    Also see the NetworkX drawing examples at
    http://networkx.lanl.gov/gallery.html

    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_edges()
    draw_networkx_labels()
    draw_networkx_edge_labels()



    """
    try:
        import matplotlib.pylab as pylab
        import numpy
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if zorder is None:
        zorder = 2

    if ax is None:
        ax=pylab.gca()

    if nodelist is None:
        nodelist=G.nodes()

    if not nodelist or len(nodelist)==0:  # empty nodelist, no drawing
        return None

    try:
        xy=numpy.asarray([pos[v] for v in nodelist])
    except KeyError as e:
        raise nx.NetworkXError('Node %s has no position.'%e)
    except ValueError:
        raise nx.NetworkXError('Bad value in node positions.')

    syms =  { # a dict from symbol to (numsides, angle)
            's' : (4,math.pi/4.0,0),   # square
            'o' : (0,0,3),            # circle
            '^' : (3,0,0),             # triangle up
            '>' : (3,math.pi/2.0,0),   # triangle right
            'v' : (3,math.pi,0),       # triangle down
            '<' : (3,3*math.pi/2.0,0), # triangle left
            'd' : (4,0,0),             # diamond
            'p' : (5,0,0),             # pentagram
            'h' : (6,0,0),             # hexagon
            '8' : (8,0,0),             # octagon
            '+' : (4,0,0),             # plus
            'x' : (4,math.pi/4.0,0)    # cross
            }

    temp_x = [p[0] for p in list(pos.values())]
    temp_y = [p[1] for p in list(pos.values())]
    minx = np.amin(temp_x)
    maxx = np.amax(temp_x)
    miny = np.amin(temp_y)
    maxy = np.amax(temp_y)

    w = max(maxx-minx,1.0)
    h = max(maxy-miny,1.0)
    #for scaling

    area2radius = lambda a: math.sqrt((a*w*h)/(ax.figure.get_figheight()*ax.figure.get_figwidth()*ax.figure.dpi*ax.figure.dpi*math.pi*.75*.75))

    if cb.iterable(node_size):
        try:
            vals = list(node_size.values())
        except:
            vals = node_size
        node_size = dict(list(zip(nodelist,list(map(area2radius,vals)))))
    else:
        node_size = {}.fromkeys(nodelist,area2radius(node_size))
    for n in node_size:
        if node_size[n] == 0.0:
            node_size[n] = .00001

    if cmap is None:
        cmap = cm.get_cmap(mpl.rcParams['image.cmap'])

    n_colors = get_color_dict(node_color,nodelist,vmin,vmax,cmap)

    sym = syms[node_shape]
    numsides,rotation,symstyle=syms[node_shape]

    node_patches = {}
    for n in nodelist:
        if symstyle==0:
            node_patches[n] = patches.RegularPolygon(pos[n],
                                                     numsides,
                                                     orientation=rotation,
                                                     radius=node_size[n],
                                                     facecolor=n_colors[n],
                                                     edgecolor='k',
                                                     alpha=alpha,
                                                     linewidth=linewidth,
                                                     transform=ax.transData,
                                                     zorder=zorder)


        elif symstyle==3:
            node_patches[n] = patches.Circle(pos[n],
                                             radius=node_size[n],
                                             facecolor=n_colors[n],
                                             edgecolor='k',
                                             alpha=alpha,
                                             linewidth=linewidth,
                                             transform=ax.transData,
                                             zorder=zorder)
        ax.add_patch(node_patches[n])




    # the pad is a little hack to deal with the fact that we don't
    # want to transform all the symbols whose scales are in points
    # to data coords to get the exact bounding box for efficiency
    # reasons.  It can be done right if this is deemed important
    temp_x = xy[:,0]
    temp_y = xy[:,1]
    minx = np.amin(temp_x)
    maxx = np.amax(temp_x)
    miny = np.amin(temp_y)
    maxy = np.amax(temp_y)

    w = maxx-minx
    h = maxy-miny
    padx, pady = 0.05*w, 0.05*h
    corners = (minx-padx, miny-pady), (maxx+padx, maxy+pady)
    ax.update_datalim(corners)
#    ax.autoscale()
    ax.autoscale_view()
    ax.set_aspect('equal')
#   pylab.axes(ax)
    #pylab.sci(node_collection)
    #node_collection.set_zorder(2)
    return node_patches

def draw_networkx_edges(G, pos, node_patches=None,
                        edgelist=None,
                        width=None,
                        edge_color=None,
                        style='solid',
                        alpha=None,
                        edge_cmap=None,
                        edge_vmin=None,
                        edge_vmax=None,
                        ax=None,
                        arrows=True,
                        arrow_style=None,
                        connection_style='arc3',
                        color_weights=False,
                        width_weights=False,
                        **kwds):
    """Draw the edges of the graph G

    This draws only the edges of the graph G.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       If not specified a spring layout positioning will be computed.
       See networkx.layout for functions that compute node positions.

    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.

    alpha: float
       The edge transparency (default=1.0)

    width`: float
       Line width of edges (default =1.0)

    edge_color: color string, or array of floats
       Edge color. Can be a single color format string (default='r'),
       or a sequence of colors with the same length as edgelist.
       If numeric values are specified they will be mapped to
       colors using the edge_cmap and edge_vmin,edge_vmax parameters.

    edge_ cmap: Matplotlib colormap
       Colormap for mapping intensities of edges (default=None)

    edge_vmin,edge_vmax: floats
       Minimum and maximum for edge colormap scaling (default=None)

    style: string
       Edge line style (default='solid') (solid|dashed|dotted,dashdot)
    arrow: Bool
       Whether to draw arrows or not for directed graphs
    arrow_style: string
       Arrow style used by matplotlib see FancyArrowPatch
    connection_style: string
       Connection style used by matplotlib, see FancyArrowPatch
    color_weights: Bool
       Whether to color the edges of a graph by their weight if the
       graph has any.
    width_weights: Bool
       Whether to vary the thicknes of an edge by their weight, if the
       graph has any.

    Examples
    --------
    >>> G=nx.dodecahedral_graph()
    >>> edges=nx.draw_networkx_edges(G,pos=nx.spring_layout(G))

    Also see the NetworkX drawing examples at
    http://networkx.lanl.gov/gallery.html

    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_nodes()
    draw_networkx_labels()
    draw_networkx_edge_labels()

    """
    try:
        import matplotlib
        import matplotlib.pylab as pylab
        import matplotlib.cbook as cb
        from matplotlib.colors import colorConverter,Colormap
        from matplotlib.collections import LineCollection
        import numpy
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if ax is None:
        ax=pylab.gca()

    if edgelist is None:
        edgelist=G.edges()

    if not edgelist or len(edgelist)==0: # no edges!
        return None

    # set edge positions
    edge_pos=numpy.asarray([(pos[e[0]],pos[e[1]]) for e in edgelist])


    if width is None and width_weights and is_weighted(G):
        lw = edge_width_weight(G,edgelist)
        if alpha is None:
            alpha = 0.75
    elif width is None:
        lw = {}.fromkeys(edgelist,1.0)
    elif cb.iterable(width):
        try:
            lwvals = list(width.values())
        except:
            lwvals = width
        lw = dict(list(zip(edgelist,lwvals)))
    elif cb.is_scalar(width):
        lw = {}.fromkeys(edgelist,width)
    else:
        raise nx.NetworkXError("Must provide a single scalar value or a list \
                                of values for line width or None")


    if edge_cmap is None:
        edge_cmap = cm.get_cmap(mpl.rcParams['image.cmap'])

    if edge_color is None and color_weights and is_weighted(G):
        edge_color = edge_color_weight(G,edgelist)
        if alpha is None:
            alpha = 0.75
    elif edge_color is None:
        edge_color = 'k'

    e_colors = get_color_dict(edge_color,edgelist,edge_vmin,edge_vmax,edge_cmap)

    edge_patches = {}

    if arrow_style is None:
        if G.is_directed():
            arrow_style = '-|>'
        else:
            arrow_style = '-'

    if node_patches is None:
        node_patches = {}.fromkeys(G.nodes(),None)
    for (u,v) in edgelist:
        edge_patches[(u,v)] = patches.FancyArrowPatch(posA=pos[u],
                                                      posB=pos[v],
                                                      arrowstyle=arrow_style,
                                                      connectionstyle=connection_style,
                                                      patchA=node_patches[u],
                                                      patchB=node_patches[v],
                                                      shrinkA=0.0,
                                                      shrinkB=0.0,
                                                      mutation_scale=20.0,
                                                      alpha=alpha,
                                                      color=e_colors[(u,v)],
                                                      lw = lw[(u,v)],
                                                      linestyle=style,
                                                      zorder=1)
        ax.add_patch(edge_patches[(u,v)])

    # update view
    minx = numpy.amin(numpy.ravel(edge_pos[:,:,0]))
    maxx = numpy.amax(numpy.ravel(edge_pos[:,:,0]))
    miny = numpy.amin(numpy.ravel(edge_pos[:,:,1]))
    maxy = numpy.amax(numpy.ravel(edge_pos[:,:,1]))

    w = maxx-minx
    h = maxy-miny
    padx, pady = 0.05*w, 0.05*h
    corners = (minx-padx, miny-pady), (maxx+padx, maxy+pady)
    ax.update_datalim( corners)
    ax.autoscale_view()

    return edge_patches



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

time_clustering.py
~~~~~~~~~~~~~~~~~~

Functions to cluster data along the time dimension.

"""

import numpy as np
import pandas as pd
import xarray as xr

try:
    import matplotlib.pyplot as plt
    import matplotlib.patches
    from matplotlib.colors import rgb2hex
    import seaborn as sns
except ImportError:
    pass  # This is logged in analysis.py

import scipy.cluster.vq as vq
from scipy.cluster import hierarchy
from scipy.spatial.distance import pdist


def _get_y_coord(array):
    if 'y' in array.coords:
        y = 'y'
    else:
        try:  # assumes a single y_ coord in array
            y = [k for k in array.coords if 'y_' in k][0]
        except IndexError:  # empty list
            y = None
    return y


def _get_datavars(data):
    return [var for var in data.data_vars if not var.startswith('_')]


def _get_timesteps_per_day(data):
    timesteps_per_day = data.attrs['time_res'] * 24
    if isinstance(timesteps_per_day, float):
        assert timesteps_per_day.is_integer(), 'Timesteps/day must be integer.'
        timesteps_per_day = int(timesteps_per_day)
    return timesteps_per_day


def reshape_for_clustering(data, tech=None):
    y_coord = 'y_def_r'
    y_values = list(data.attrs['_sets'][y_coord])
    timesteps_per_day = _get_timesteps_per_day(data)
    days = int(len(data.t) / timesteps_per_day)
    regions = len(data['x'])
    techs = len(y_values)

    if tech is None:
        X = (data.r.loc[{'y': y_values}]
                   .transpose('t', 'x', 'y')
                   .values.reshape(days, timesteps_per_day * regions * techs))
    else:
        X = (data.r.loc[{'y': tech}]
                   .values
                   .reshape((days, timesteps_per_day * regions)))

    return np.nan_to_num(X)  # replace any NaN with 0


def reshape_clustered(clustered, data, tech=None):
    y_coord = 'y_def_r'
    y_values = list(data.attrs['_sets'][y_coord])
    timesteps_per_day = _get_timesteps_per_day(data)
    days = clustered.shape[0]
    regions = len(data['x'])
    techs = len(y_values)

    if tech is None:
        reshaped = clustered.reshape(techs, days * timesteps_per_day, regions)
    else:
        reshaped = clustered.reshape(1, days * timesteps_per_day, regions)

    arr = xr.DataArray(reshaped, dims=('y', 't', 'x'))

    arr['x'] = data['x']
    if tech is None:
        arr['y'] = y_values
    else:
        arr['y'] = tech

    return arr


def get_mean_from_clusters(data, clusters, timesteps_per_day):

    cluster_map = clusters.groupby(clusters).groups

    hour_of_day = pd.Series([pd.Timestamp(i).hour
                             for i in data.coords['t'].values])
    # hour_of_day = pd.Series([i - (i // timesteps_per_day) * timesteps_per_day
    #                          for i in data.coords['t'].values])

    ds = {}
    data_vars_in_t = [v for v in _get_datavars(data)
                      if 't' in data[v].dims]
    for var in data_vars_in_t:
        data_arrays = []
        array = data[var]
        y_coord = _get_y_coord(array)

        t_coords = ['{}-{}'.format(cid, t)
                    for cid in cluster_map
                    for t in range(timesteps_per_day)]

        # Check if this variable has any data
        if len(data.coords[y_coord].values) == 0:
            e = np.empty((0, len(t_coords), len(data.coords['x'].values)))
            ds[var] = xr.DataArray(e, dims=[y_coord, 't', 'x'],
                                   coords={'x': data.coords['x'].values,
                                           't': t_coords})
        else:
            for cluster_id, cluster_members in cluster_map.items():
                y_arrays = []
                for y in data.coords[y_coord].values:
                    d = (array.loc[{'t': cluster_members, y_coord: y}]
                              .to_pandas().groupby(lambda x: x.hour).mean()
                              .to_xarray().to_array(dim='x').T
                         )
                    d.coords[y_coord] = y
                    d = d.rename({'index': 't'})
                    d.coords['t'] = [i for i in t_coords
                                     if i.startswith('{}-'.format(cluster_id))]
                    y_arrays.append(d)
                data_arrays.append(xr.concat(y_arrays, dim=y_coord))
                # data_arrays[-1].coords['cluster'] = cluster_id

            ds[var] = xr.concat(data_arrays, dim='t')
    ds = xr.Dataset(ds)
    return ds


def find_nearest_vector_index(array, value):
    return np.array([np.linalg.norm(x + y + z + u)
                     for (x, y, z, u) in array - value]).argmin()


def get_closest_days_from_clusters(data, mean_data, clusters):
    subset_y = list(data.attrs['_sets']['y_def_r'])
    dtindex = data['t'].to_index()
    ts_per_day = _get_timesteps_per_day(data)
    days = int(len(data['t']) / ts_per_day)
    n_x = len(data['x'])
    n_y = len(subset_y)

    chosen_days = {}

    for cluster in sorted(clusters.unique()):

        subset_t = [t for t in mean_data.t.values if t.startswith('{}-'.format(cluster))]

        target = mean_data['r'].loc[dict(t=subset_t, y=subset_y)].values

        lookup_array = data['r'].loc[dict(y=subset_y)].values
        lookup_array = lookup_array.reshape((n_y, days, ts_per_day, n_x)).transpose(1, 0, 2, 3)

        chosen_days[cluster] = find_nearest_vector_index(lookup_array, target)

    days_list = sorted(list(set(chosen_days.values())))
    new_t_coord = _hourly_from_daily_index(dtindex[::ts_per_day][days_list])

    chosen_day_timestamps = {k: dtindex[::ts_per_day][v]
                             for k, v in chosen_days.items()}

    new_data = data.loc[dict(t=new_t_coord)]

    return new_data, chosen_day_timestamps


def _hourly_from_daily_index(idx):
    dtrange = lambda i: pd.date_range(i, i + pd.Timedelta('1D'),
                                      freq='1H')[:-1]
    new_idx = pd.concat([pd.Series(1, dtrange(i))
                         for i in idx]).index
    return new_idx


def map_clusters_to_data(data, clusters, how):
    """
    Returns a copy of data that has been clustered.

    Parameters
    ----------
    how : str
        How to select data from clusters.
        Can be mean (centroid) or closest.

    """
    # FIXME hardcoded time intervals ('1H', '1D')

    # Get all timesteps, not just the first per day
    ts_per_day = _get_timesteps_per_day(data)
    idx = clusters.index
    new_idx = _hourly_from_daily_index(idx)
    clusters_timeseries = (clusters.reindex(new_idx)
                           .fillna(method='ffill').astype(int))

    new_data = get_mean_from_clusters(data, clusters_timeseries, ts_per_day)

    if how == 'mean':
        # Add timestep names by taking the median timestamp from daily clusters...
        # (a random way of doing it, but we want some label to apply)
        timestamps = clusters.groupby(clusters).apply(lambda x: x.index[int(len(x.index) / 2)])
        new_t_coord = pd.concat([pd.Series(pd.date_range(ts, ts + pd.Timedelta('1D'), freq='1H')[:-1])
                                for ts in timestamps], ignore_index=True)
        new_data.coords['t'] = new_t_coord.as_matrix()

        # Generate weights
        # weight of each timestep = number of timesteps in this timestep's cluster
        # divided by timesteps per day (since we're grouping days together and
        # a cluster consisting of 1 day = 24 hours should have weight of 1)
        value_counts = clusters_timeseries.value_counts() / ts_per_day
        # And turn the index into dates (days)
        value_counts = pd.DataFrame({
            'dates': timestamps,
            'counts': value_counts}).set_index('dates')['counts']

    elif how == 'closest':
        new_data, chosen_ts = get_closest_days_from_clusters(data, new_data, clusters)
        # Deal with the case where more than one cluster has the same closest day
        # An easy way is to rename the original clusters with the chosen days
        # So at this point, clusterdays_timeseries maps all timesteps to the day
        # of year of the cluster the timestep belongs to
        clusterdays_timeseries = clusters_timeseries.map(lambda x: chosen_ts[x])
        value_counts = clusterdays_timeseries.value_counts() / ts_per_day

    weights = (value_counts.reindex(_hourly_from_daily_index(value_counts.index))
                           .fillna(method='ffill'))
    new_data['_weights'] = xr.DataArray(weights, dims=['t'])
    new_data['_time_res'] = xr.DataArray(np.ones(len(new_data['t'])) * (24 / ts_per_day),
                                         coords={'t': new_data['t']})

    return new_data


def get_clusters_kmeans(data, tech=None, timesteps=None, k=5):
    """
    Parameters
    ----------
    data : xarray.Dataset
        Should be normalized

    Returns
    -------
    clusters : dataframe
        Indexed by timesteps and with locations as columns, giving cluster
        membership for first timestep of each day.
    centroids

    """
    timesteps_per_day = _get_timesteps_per_day(data)

    if timesteps is not None:
        data = data.loc[{'t': timesteps}]
    else:
        timesteps = data.t.values

    X = reshape_for_clustering(data, tech)

    centroids, distortion = vq.kmeans(X, k)

    # Determine the cluster membership of each day
    day_clusters = vq.vq(X, centroids)[0]

    # Create mapping of timesteps to clusters
    clusters = pd.Series(day_clusters, index=timesteps[::timesteps_per_day])

    # Reshape centroids
    centroids = reshape_clustered(centroids, data, tech)

    return clusters, centroids


def get_clusters_hierarchical(data, tech=None, max_d=None, k=None):
    """
    Parameters
    ----------
    data : xarray.Dataset
        Should be normalized
    max_d : float or int, optional
        Max distance for returning clusters.
    k : int, optional
        Number of desired clusters.

    Returns
    -------
    clusters
    X
    Z

    """
    X = reshape_for_clustering(data, tech)

    # Generate the linkage matrix
    Z = hierarchy.linkage(X, 'ward')

    if max_d:
        # Get clusters based on maximum distance
        clusters = hierarchy.fcluster(Z, max_d, criterion='distance')
    elif k:
        # Get clusters based on number of desired clusters
        clusters = hierarchy.fcluster(Z, k, criterion='maxclust')
    else:
        clusters = None

    # Make sure clusters are a pd.Series with a datetime index
    if clusters is not None:
        timesteps_per_day = _get_timesteps_per_day(data)
        timesteps = data.coords['t'].values  # All timesteps

        clusters = pd.Series(clusters, index=timesteps[::timesteps_per_day])

    return (clusters, X, Z)


def fancy_dendrogram(*args, **kwargs):
    """
    Code adapted from:
    https://joernhees.de/blog/2015/08/26/scipy-hierarchical-clustering-and-dendrogram-tutorial/

    """
    max_d = kwargs.pop('max_d', None)
    if max_d and 'color_threshold' not in kwargs:
        kwargs['color_threshold'] = max_d
    annotate_above = kwargs.pop('annotate_above', 0)

    ddata = hierarchy.dendrogram(*args, **kwargs)

    if not kwargs.get('no_plot', False):
        plt.xlabel('Sample index or (cluster size)')
        plt.ylabel('Distance')
        for i, d, c in zip(ddata['icoord'], ddata['dcoord'], ddata['color_list']):
            x = 0.5 * sum(i[1:3])
            y = d[1]
            if y > annotate_above:
                plt.plot(x, y, 'o', c=c)
                plt.annotate("%.3g" % y, (x, y), xytext=(0, -5),
                             textcoords='offset points',
                             va='top', ha='center', fontsize=10)
        if max_d:
            plt.axhline(y=max_d, c='grey', lw=1.0, alpha=0.5)
    return ddata


def draw_dendrogram(Z, max_d):
    fig = plt.figure(figsize=(12, 6))
    ax = plt.subplot(111)

    # link_color_pal = sns.color_palette("hls", 8)
    link_color_pal = sns.color_palette("Set2", 10)

    hierarchy.set_link_color_palette([rgb2hex(i) for i in link_color_pal])

    color_threshold = max_d

    fancy_dendrogram(
        Z,
        ax=ax,
        color_threshold=color_threshold,
        truncate_mode='lastp',
        p=20,
        leaf_rotation=90,
        leaf_font_size=10,
        show_contracted=True,
        annotate_above=5,  # useful in small plots so annotations don't overlap
        max_d=color_threshold,
    )

    sns.despine()

    # Modify the contracted markers
    for child in ax.get_children():
        if isinstance(child, matplotlib.patches.Ellipse):
            child.set_zorder(1000)
            child.set_alpha(0.3)

    return fig


def cophenetic_corr(X, Z):
    """
    Get the Cophenetic Correlation Coefficient of a clustering with help
    of the cophenet() function. This (very very briefly) compares (correlates)
    the actual pairwise distances of all your samples to those implied by the
    hierarchical clustering. The closer the value is to 1, the better
    the clustering preserves the original distances.

    Source:
    https://joernhees.de/blog/2015/08/26/scipy-hierarchical-clustering-and-dendrogram-tutorial/

    """
    c, coph_dists = hierarchy.cophenet(Z, pdist(X))
    return c, coph_dists



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

analysis.py
~~~~~~~~~~~

Functionality to analyze model results.

"""

import logging

try:
    from matplotlib.patches import Rectangle
    from matplotlib.colors import ListedColormap
except ImportError:
    logging.debug('Matplotlib could not be imported, '
                  'no plotting will be available.')
import numpy as np
import pandas as pd

from . import core
from . import utils
from . import analysis_utils as au


def plot_carrier_production(solution, carrier='power', subset=dict(),
                            **kwargs):
    """
    Generate a stackplot of the production by the given ``carrier``.

    Parameters
    ----------
    solution : model solution xarray.Dataset
    carrier : str, optional
        Name of the carrier to plot, default 'power'.
    subset : dict, optional
        Specify an additional subset of Dataset coordinates, for example,
        dict(t=slice('2005-02-01', '2005-02-10').
    **kwargs : optional
        Passed to ``plot_timeseries``.

    """
    data = solution['e'].loc[dict(c=carrier, **subset)].sum(dim='x')
    return plot_timeseries(solution, data, carrier=carrier, demand='demand_{}'.format(carrier), **kwargs)


def plot_timeseries(solution, data, carrier='power', demand='demand_power',
                    types=['supply', 'conversion', 'storage', 'unmet_demand'],
                    colormap=None, ticks=None,
                    resample_options=None, resample_func=None,
                    add_legend=True, ax=None):
    """
    Generate a stackplot of ``data`` for the given ``carrier``,
    plotting ``demand`` on top.

    Use ``plot_carrier_production`` for a simpler way to plot production by
    a given carrier.

    Parameters
    ----------
    solution : model solution xarray.Dataset
    data : xarray.Dataset
        Subset of solution to plot.
    carrier : str, optional
        Name of the carrier to plot, default 'power'.
    demand : str, optional
        Name of a demand tech whose time series to plot on top,
        default 'demand_power'.
    types : list, optional
        Technology types to include in the plot. Default list is
        ['supply', 'conversion', 'storage', 'unmet_demand'].
    colormap : matplotlib colormap, optional
        Colormap to use. If not given, the colors specified for each
        technology in the solution's metadata are used.
    ticks : str, optional
        Where to draw x-axis (time axis) ticks. By default (None),
        auto-detects, but can manually set to either 'hourly', 'daily',
        or 'monthly'.
    resample_options : dict, optional
        Give options for pandas.DataFrame.resample in a dict, to resample
        the entire time series prior to plotting. Both resample_options
        and resample_func must be given for resampling to happen.
        Default None.
    resample_func : string, optional
        Give the name of the aggregating function to use when resampling,
        e.g. "mean" or "sum". Default None.

    """
    # Determine ticks
    if not ticks:
        days = (data.coords['t'].values[-1] - data.coords['t'].values[0])
        timespan = days / np.timedelta64(1, 'D')
        # timespan = days.astype('timedelta64[D]').astype(int)
        if timespan <= 2:
            ticks = 'hourly'
        elif timespan < 14:
            ticks = 'daily'
        else:
            ticks = 'monthly'
    # Set up time series to plot, dividing it by time_res_series
    time_res = solution['time_res'].to_pandas()
    data = data.to_pandas().T
    plot_df = data.divide(time_res, axis='index')
    if resample_options and resample_func:
        plot_df = getattr(plot_df.resample(**resample_options), resample_func)()
    # Get tech stack and names
    df = solution['metadata'].to_pandas().query('carrier == "{}"'.format(carrier))
    query_string = au._get_query_string(types)
    stacked_techs = df.query(query_string).index.tolist()
    # Put stack in order according to stack_weights
    weighted = df.stack_weight.sort_values(ascending=False).index.tolist()
    stacked_techs = [y for y in weighted if y in stacked_techs]
    names = [df.at[y, 'name'] for y in stacked_techs]
    # If no colormap given, derive one from colors given in metadata
    if not colormap:
        colors = [df.at[i, 'color'] for i in stacked_techs]
        colormap = ListedColormap(colors)
    # Plot!
    if ax is None:
        ax = au.stack_plot(plot_df, stacked_techs, colormap=colormap,
                           alpha=0.9, ticks=ticks, legend=None, names=names)
    else:
        au.stack_plot(plot_df, stacked_techs, colormap=colormap,
                      alpha=0.9, ticks=ticks, legend=None, names=names,
                      ax=ax)
    ax.plot(plot_df[demand].index,
            plot_df[demand] * -1,
            color='red', lw=1.5, ls='--', label=df.at[demand, 'name'])
    # Add legend here rather than in stack_plot so we get demand too
    if add_legend:
        au.legend_outside_ax(ax, where='right')
    return ax


def plot_installed_capacities(solution,
                              types=['supply', 'conversion', 'storage'],
                              unit_multiplier=1.0,
                              unit_label='kW',
                              **kwargs):
    """
    Plot installed capacities (``e_cap``) with a bar plot.

    Parameters
    ----------
    types : list, default ['supply', 'conversion', 'storage']
        Technology types to include in the plot.
    unit_multiplier : float or int, default 1.0
        Multiply installed capacities by this value for plotting.
    unit_label : str, default 'kW'
        Label for capacity values, adjust this when
        changing ``unit_multiplier``.
    **kwargs : are passed to ``pandas.DataFrame.plot()``

    """
    query_string = au._get_query_string(types)
    md = solution.metadata.to_pandas()
    supply_cap = md.query(query_string).index.tolist()

    df = solution['e_cap'].loc[dict(y=supply_cap)].to_pandas()

    weighted = md.stack_weight.sort_values(ascending=False).index.tolist()
    stacked_techs = [y for y in weighted if y in df.columns]

    df = df.loc[:, stacked_techs] * unit_multiplier

    names = [md.at[y, 'name'] for y in df.columns]
    colors = [md.at[i, 'color'] for i in df.columns]
    colormap = ListedColormap(colors)
    proxies = [Rectangle((0, 0), 1, 1, fc=i)
               for i in colors]

    # Order the locations nicely, but only take those locations that actually
    # exists in the current solution
    if ('metadata' in solution.config_model and
            'location_ordering' in solution.config_model.metadata):
        meta_config = solution.config_model.metadata
        for index, item in enumerate(meta_config.location_ordering):
            if item in df.index:
                df.at[item, 'ordering'] = index
        df = df.sort_values(by='ordering', ascending=False)
        df = df.drop('ordering', axis=1)

    ax = df.plot(kind='barh', stacked=True, legend=False, colormap=colormap,
                 **kwargs)
    leg = au.legend_outside_ax(ax, where='right', artists=proxies, labels=names)

    ylab = ax.set_ylabel('')
    xlab = ax.set_xlabel('Installed capacity ({})'.format(unit_label))

    return ax


def plot_transmission(solution, tech='hvac', carrier='power',
                      labels='utilization',
                      figsize=(15, 15), fontsize=9,
                      show_scale=True,
                      ax=None,
                      **kwargs):
    """
    Plot transmission links on a map. Requires that model metadata have
    been defined with a lat/lon for each model location and a boundary for
    the map display.

    Requires Basemap and NetworkX to be installed.

    Parameters
    ----------
    solution : solution container
    tech : str, default 'hvac'
        Which transmission technology to plot.
    carrier : str, default 'power'
        Which carrier to plot transmission for.
    labels : str, default 'utilization'
        Determines how transmission links are labeled, either
        `transmission` or `utilization`.
    figsize : (int, int), default (15, 15)
        Size of resulting figure.
    fontsize : int, default 9
        Font size of figure labels.
    show_scale : bool, default True
        Plot a distance scale on the map.
    ax : matplotlib axes, default None
    **kwargs : are passed to ``analysis_utils.plot_graph_on_map()``

    """
    import networkx as nx

    # Determine maximum that could have been transmitted across a link
    def get_edge_capacity(solution, a, b):
        hrs = solution['time_res'].to_pandas().sum()
        cap = solution['e_cap_net'].loc[dict(x=a, y='{}:'.format(tech) + b)].to_pandas() * hrs
        return cap

    # Get annual power transmission between zones
    zones = sorted(list(solution.coords['x'].values))
    trans_tech = lambda x: '{}:{}'.format(tech, x)
    def get_annual_power_transmission(zone):
        try:
            return solution['e'].loc[dict(c=carrier, y=trans_tech(zone))].sum(dim='t').to_pandas()
        except KeyError:
            return 0
    df = pd.DataFrame({zone: get_annual_power_transmission(zone)
                      for zone in zones}, index=zones)

    # Set negative and very small (erroneous) values to zero --
    # Positive values = energy received at a node by transmission,
    # negative values = energy sent by transmission.
    # The absolute value sent and recieved between two nodes will be
    # different when there is line loss (i.e. negative values will have a
    # larger absolute value than the positive value for the same transmission
    # line), which is why we take the recieving (positive) value.
    df[df < df.max().max()*1e-10] = 0

    # Create directed graph
    G = nx.from_numpy_matrix(df.as_matrix().T, create_using=nx.DiGraph())
    G = nx.relabel_nodes(G, dict(list(zip(list(range(len(zones))), zones))))

    # Transmission
    edge_transmission = {edge: int(round(df.at[edge[1], edge[0]] / 1e6))
                         for edge in G.edges()}

    # Utilization ratio
    edge_use = {(a, b): (df.at[a, b] + df.at[b, a])
                / get_edge_capacity(solution, a, b)
                for (a, b) in G.edges()}

    # Set edge labels
    if labels == 'utilization':
        edge_labels = {k: '{:.2f}'.format(v) for k, v in edge_use.items()}
    elif labels == 'transmission':
        edge_labels = edge_transmission

    # Set edge colors
    edge_colors = [edge_use[i] for i in G.edges()]

    ax, m = au.plot_graph_on_map(solution.config_model, G=G,
                                 edge_colors=edge_colors,
                                 edge_labels=edge_labels,
                                 figsize=figsize, fontsize=fontsize,
                                 ax=ax, show_scale=show_scale, **kwargs)

    return ax


def get_delivered_cost(solution, cost_class='monetary', carrier='power',
                       count_unmet_demand=False):
    """
    Get the levelized cost per unit of energy delivered for the given
    ``cost_class`` and ``carrier``.

    Parameters
    ----------
    solution : solution container
    cost_class : str, default 'monetary'
    carrier : str, default 'power'
    count_unmet_demand : bool, default False
        Whether to count the cost of unmet demand in the final
        delivered cost.

    """
    summary = solution.summary.to_pandas()
    meta = solution.metadata.to_pandas()
    carrier_subset = meta[meta.carrier == carrier].index.tolist()
    if count_unmet_demand is False:
        try:
            carrier_subset.remove('unmet_demand_' + carrier)
        except ValueError:  # no unmet demand technology
            pass
    cost = solution['costs'].loc[dict(k=cost_class, y=carrier_subset)].to_pandas().sum().sum()
    # Actually, met_demand also includes demand "met" by unmet_demand
    met_demand = summary.at['demand_' + carrier, 'e_con']
    try:
        unmet_demand = summary.at['unmet_demand_' + carrier, 'e_con']
    except KeyError:
        unmet_demand = 0
    if count_unmet_demand is False:
        demand = met_demand + unmet_demand  # unmet_demand is positive, add it
    else:
        demand = met_demand

    return cost / demand * -1


def get_levelized_cost(solution, cost_class='monetary', carrier='power',
                       group=None, locations=None,
                       unit_multiplier=1.0):
    """
    Get the levelized cost per unit of energy produced for the given
    ``cost_class`` and ``carrier``, optionally for a subset of technologies
    given by ``group`` and a subset of ``locations``.

    Parameters
    ----------
    solution : solution container
    cost_class : str, default 'monetary'
    carrier : str, default 'power'
    group : str, default None
        Limit the computation to members of the given group (see the
        groups table in the solution for valid groups).
    locations : str or iterable, default None
        Limit the computation to the given location or locations.
    unit_multiplier : float or int, default 1.0
        Adjust unit of the returned cost value. For example, if model units
        are kW and kWh, ``unit_multiplier=1.0`` will return cost per kWh, and
        ``unit_multiplier=0.001`` will return cost per MWh.

    """
    if group is None:
        group = 'supply'

    members = solution.groups.to_pandas().at[group, 'members'].split('|')

    if locations is None:
        locations_slice = slice(None)
    elif isinstance(locations, (str, float, int)):
        # Make sure that locations is a list if it's a single value
        locations_slice = [locations]
    else:
        locations_slice = locations

    cost = solution['costs'].loc[dict(k=cost_class, x=locations_slice, y=members)]
    ec_prod = solution['ec_prod'].loc[dict(c=carrier, x=locations_slice, y=members)]

    if locations is None:
        cost = cost.sum(dim='x').to_pandas()
        ec_prod = ec_prod.sum(dim='x').to_pandas()
    else:
        cost = cost.to_pandas()
        ec_prod = ec_prod.to_pandas()

    return (cost / ec_prod) * unit_multiplier


def get_group_share(solution, techs, group_type='supply',
                    var='e_prod'):
    """
    From ``solution.summary``, get the share of the given list of ``techs``
    from the total for the given ``group_type``, for the given ``var``.

    """
    summary = solution.summary
    meta = solution.metadata.to_pandas()
    group = meta.query('type == "' + group_type + '"').index.tolist()
    if group_type == 'transmission':
        # Special case for transmission techs, we only want
        # the base tech names, since we're looking this
        # up in the summary table
        def transmission_basenames(group):
            return list(set([i.split(':')[0] for i in group]))
        techs = transmission_basenames(techs)
        group = transmission_basenames(group)
    supply_total = summary.loc[group, var].sum()
    supply_group = summary.loc[techs, var].sum()
    try:
        with np.errstate(divide='ignore', invalid='ignore'):
            return supply_group / supply_total
    except ZeroDivisionError:
        # FIXME it seems that on some systems, supply_ are not numpy.floats
        # but regular Python floats, leading to a ZeroDivisionError
        # in cases such as demand groups
        return np.nan


def get_unmet_demand_hours(solution, carrier='power', details=False):
    """
    Get information about unmet demand from ``solution``.

    Parameters
    ----------
    solution : solution container
    carrier : str, default 'power'
    details : bool, default False
        By default, only the number of hours with unmet are returned. If
        details is True, a dict with 'hours', 'timesteps', and 'dates' keys
        is returned instead.

    """
    unmet = (solution['e']
             .loc[dict(c=carrier, y='unmet_demand_' + carrier)]
             .sum(dim='x')
             .to_pandas())
    timesteps = len(unmet[unmet > 0])
    hours = solution.time_res.to_pandas()[unmet > 0].sum()
    if details:
        return {'hours': hours, 'timesteps': timesteps,
                'dates': unmet[unmet > 0].index}
    else:
        return hours


def areas_below_resolution(solution, resolution):
    """
    Returns a list of (start, end) timestamp tuples delimiting those
    areas in the solution below the given timestep resolution (in hours).

    """
    # TODO: add unit tests
    time_res = solution.time_res.to_pandas()
    selected = time_res[time_res < resolution]
    return list(au._get_ranges(selected.index.tolist()))


def get_swi(solution, shares_var='e_cap', exclude_patterns=['unmet_demand']):
    """
    Returns the Shannon-Wiener diversity index.

    :math:`SWI = -1 \\times \sum_{i=1}^{I} p_{i} \\times \ln(p_{i})`

    where where I is the number of categories and :math:`p_{i}`
    is each category's share of the total (between 0 and 1).

    :math:`SWI` is zero when there is perfect concentration.

    """
    # TODO: add unit tests
    techs = au.get_supply_groups(solution)
    for pattern in exclude_patterns:
        techs = [t for t in techs if pattern not in t]
    swi = -1 * sum((p * np.log(p))
                   for p in [solution.shares.to_pandas().at[y, shares_var]
                   for y in techs]
                   if p > 0)
    return swi


def get_hhi(solution, shares_var='e_cap', exclude_patterns=['unmet_demand']):
    """
    Returns the Herfindahl-Hirschmann diversity index.

    :math:`HHI = \sum_{i=1}^{I} p_{i}^2`

    where :math:`p_{i}` is the percentage share of each technology i (0-100).

    :math:`HHI` ranges between 0 and 10,000. A value above 1800 is
    considered a sign of a concentrated market.

    """
    # TODO: add unit tests
    techs = au.get_supply_groups(solution)
    for pattern in exclude_patterns:
        techs = [t for t in techs if pattern not in t]
    hhi = sum((solution.shares.to_pandas().at[y, shares_var] * 100.) ** 2
              for y in techs)
    return hhi


def get_domestic_supply_index(solution):
    """
    Assuming that ``solution`` specifies a ``domestic`` cost class to
    give each technology a domesticity score, return the total domestic
    supply index for the given solution.

    """
    # TODO: add unit tests
    idx = solution.metadata.query('type == "supply"').index.tolist()
    dom = (solution.costs.loc[dict(k='domestic', y=idx)].sum().sum() /
           solution['ec_prod'].loc[dict(c='power')].sum().sum())
    return dom


def map_results(results, func, as_frame=False):
    """
    Applies ``func`` to each model solution in ``results``, returning
    a pandas DataFrame (if as_frame is True) or Series,
    indexed by the run names (if available).

    """
    # TODO: add unit tests
    def map_func(x):
        try:
            return func(x)
        except Exception:
            return np.nan

    iterations = list(results.solutions.keys())
    items = [map_func(results.solutions[i]) for i in iterations]
    idx = list(results.solutions.keys())
    try:
        idx = results.iterations.loc[idx, 'name']
    except Exception:
        pass
    if as_frame:
        return pd.DataFrame(items, index=idx)
    else:
        return pd.Series(items, index=idx)


class SolutionModel(core.BaseModel):
    def __init__(self, solution):
        """
        Dummy model created from a model solution,
        which gives access to the ``config_model`` AttrDict and a subset
        of Model methods, including ``get_option``, ``get_cost``, and
        ``get_depreciation``.

        Furthermore, it provides several ``recompute_`` methods
        to recalculate costs for a given technology after modifying the
        underlying data.

        """
        super().__init__()
        self.config_model = solution.config_model

        # Reconstruct self.data from solution
        params = ['r', 'e_eff']
        try:
            self.data = solution[params]
        except KeyError:
            pass

        # Attach solution (minus data parameters)
        for p in params:
            try:
                del solution[p]
            except KeyError:
                pass
        self.solution = solution

        # Add getters
        self._get_option = utils.option_getter(self.config_model)
        self._get_cost = utils.cost_getter(self._get_option)
        self._get_dep = utils.depreciation_getter(self._get_option)

    def get_option(self, *args, **kwargs):
        return self._get_option(*args, **kwargs)

    def get_cost(self, *args, **kwargs):
        return self._get_cost(*args, **kwargs)

    def get_depreciation(self, *args, **kwargs):
        return self._get_dep(*args, **kwargs)

    def recompute_investment_costs(self, y, k='monetary',
                                   cost_adjustment=1.0):
        """
        Recompute investment costs for the given technology ``y``.

        ``cost_adjustment`` allows linear scaling of the costs by the
        given amount.

        """
        get_cost = self.get_cost
        solution = self.solution
        cost_con = pd.DataFrame()
        cost_con['e_cap'] = (get_cost('e_cap', y, k)
                             * solution['e_cap'].loc[dict(y=y)])
        cost_con['r_cap'] = (get_cost('r_cap', y, k)
                             * solution['r_cap'].loc[dict(y=y)])
        cost_con['r_area'] = (get_cost('r_area', y, k)
                              * solution['r_area'].loc[dict(y=y)])
        cost_con['rb_cap'] = (get_cost('rb_cap', y, k)
                              * solution['rb_cap'].loc[dict(y=y)])
        cost_con['s_cap'] = (get_cost('s_cap', y, k)
                             * solution['s_cap'].loc[dict(y=y)])
        cost_con['total'] = cost_con.sum(axis=1)
        cost_con['total_per_e_cap'] = (cost_con['total']
                                       / solution['e_cap'].loc[dict(y=y)])
        cost_con.at['total', 'total'] = cost_con['total'].sum()
        total_per_e_cap = (cost_con.at['total', 'total']
                           / solution['e_cap'].loc[dict(y=y)].sum())
        cost_con.at['total', 'total_per_e_cap'] = total_per_e_cap
        return cost_con * cost_adjustment

    def recompute_operational_costs(self, y, k='monetary', carrier='power',
                                    cost_adjustment=1.0):
        """
        Recompute operational costs for the given technology ``y``.

        ``cost_adjustment`` linearly scales the construction costs used
        internally, but is not applied to the rest of the operational
        costs.

        """
        get_cost = self.get_cost
        get_depreciation = self.get_depreciation
        solution = self.solution
        # Here we want gross production
        production = solution['es_prod'].loc[dict(c=carrier, y=y)]
        fuel = (solution['rs'].loc[dict(y=y)].sum(dim='t')
                / self.get_option(y + '.constraints.r_eff'))
        fuel_rb = (solution['rbs'].loc[dict(y=y)].sum(dim='t')
                   / self.get_option(y + '.constraints.rb_eff'))
        icf = self.recompute_investment_costs
        cost_con = icf(y, k=k, cost_adjustment=cost_adjustment)['total'].sum()
        cost_op = pd.DataFrame()
        cost_op['om_frac'] = (get_cost('om_frac', y, k) * cost_con
                              * get_depreciation(y, k)
                              * (solution['time_res'].to_pandas().sum() / 8760))
        cost_op['om_fixed'] = (get_cost('om_fixed', y, k)
                               * solution['e_cap'].loc[dict(y=y)]
                               * (solution['time_res'].to_pandas().sum() / 8760))
        cost_op['om_var'] = get_cost('om_var', y, k) * production
        cost_op['om_fuel'] = get_cost('om_fuel', y, k) * fuel
        cost_op['om_rb'] = get_cost('om_rb', y, k) * fuel_rb
        cost_op['total'] = cost_op.sum(axis=1)
        return cost_op

    def recompute_levelized_costs(self, y, k='monetary', carrier='power',
                                  cost_adjustment=1.0):
        """
        Recompute levelized costs for the given technology ``y``.

        ``cost_adjustment`` linearly scales the construction costs used
        internally to compute levelized costs.

        """
        get_depreciation = self.get_depreciation
        solution = self.solution
        icf = self.recompute_investment_costs
        cost_con = icf(y, k=k, cost_adjustment=cost_adjustment)['total']
        ocf = self.recompute_operational_costs
        cost_op = ocf(y, k=k, carrier=carrier,
                      cost_adjustment=cost_adjustment)['total']
        # Here we want net production
        production = solution['ec_prod'].loc[dict(c=carrier, y=y)]
        lc = (get_depreciation(y, k) * (solution['time_res'].to_pandas().sum() / 8760)
              * cost_con + cost_op) / production.to_pandas()
        lc.at['total'] = (get_depreciation(y, k)
                          * (solution['time_res'].to_pandas().sum() / 8760)
                          * cost_con.total
                          # * cost_con.drop(['total'], axis=0).sum()
                          + cost_op.sum()) / production.sum()
        return lc



from io import StringIO

import pandas as pd
import pytest

from calliope import locations, utils


class TestLocations:
    @pytest.fixture
    def sample_locations(self):
        setup = StringIO("""
        test:
            techs: ['demand', 'unmet_demand', 'ccgt']
            override:
                demand:
                    x_map: 'a: test'
                    constraints:
                        r: file=demand.csv
                        r_scale_to_peak: -10
                ccgt:
                    constraints:
                        e_cap.max: 10
        test1:
            within: test
            techs: ['csp']
        """)
        return utils.AttrDict.from_yaml(setup)

    @pytest.fixture
    def sample_unexploded_locations(self):
        setup = StringIO("""
        1,2,3:
            foo:
        a,b,c:
            foo:
        10--15:
            foo:
        10-20:
            foo:
        21--23,25,z:
            foo:
        x:
            foo:
        y:
            foo:
        """)
        return utils.AttrDict.from_yaml(setup)

    @pytest.fixture
    def sample_nested_locations(self):
        setup = StringIO("""
        1,2,3:
            techs: ['foo']
        foo:
            techs: ['foo']
        10,11,12:
            within: 1
            techs: ['foo']
        20,21,22:
            within: 2
            techs: ['foo']
        bar,baz:
            within: foo
            techs: ['foo']
        qux:
            within: bar
            techs: ['foo']
        """)
        return utils.AttrDict.from_yaml(setup)

    @pytest.fixture
    def sample_overlapping_locations(self):
        setup = StringIO("""
        1,2,3:
            techs: ['foo']
        1:
            override:
                bar: baz
        """)
        return utils.AttrDict.from_yaml(setup)

    def test_generate_location(self, sample_locations):
        location = 'test'
        items = sample_locations[location]
        techs = ['demand', 'unmet_demand', 'ccgt']
        result = locations._generate_location(location, items, techs)
        wanted_cols = ['_level', '_location',
                       '_override.ccgt.constraints.e_cap.max',
                       '_override.demand.constraints.r',
                       '_override.demand.constraints.r_scale_to_peak',
                       '_override.demand.x_map',
                       '_within', 'ccgt', 'demand', 'unmet_demand']
        assert sorted(result.keys()) == wanted_cols

    def test_generate_location_lacking_techs(self, sample_locations):
        location = 'test'
        items = sample_locations[location]
        techs = ['unmet_demand', 'ccgt']
        result = locations._generate_location(location, items, techs)
        wanted_cols = ['_level', '_location',
                       '_override.ccgt.constraints.e_cap.max',
                       '_override.demand.constraints.r',
                       '_override.demand.constraints.r_scale_to_peak',
                       '_override.demand.x_map',
                       '_within', 'ccgt', 'unmet_demand']
        assert sorted(result.keys()) == wanted_cols

    def test_explode_location_single(self):
        assert locations.explode_locations('a') == ['a']

    def test_explode_location_range(self):
        assert locations.explode_locations('1--3') == ['1', '2', '3']

    def test_explode_location_range_backwards(self):
        with pytest.raises(KeyError):
            locations.explode_locations('3--1')

    def test_explode_location_range_nonnumeric(self):
        with pytest.raises(ValueError):
            locations.explode_locations('a--c')

    def test_explore_location_list(self):
        assert locations.explode_locations('1,2,3') == ['1', '2', '3']

    def test_explode_location_mixed(self):
        assert (locations.explode_locations('a,b,1--3,c')
                == ['a', 'b', '1', '2', '3', 'c'])

    def test_explode_location_empty(self):
        with pytest.raises(KeyError):
            assert locations.explode_locations('')

    def test_explode_location_invalid(self):
        with pytest.raises(AssertionError):
            assert locations.explode_locations(['a', 'b'])

    def test_process_locations(self, sample_unexploded_locations):
        fixture = sample_unexploded_locations
        o = locations.process_locations(fixture)
        assert '10' in o
        assert 'x' in o
        assert len(list(o.keys())) == 20

    def test_process_locations_overlap(self, sample_overlapping_locations):
        fixture = sample_overlapping_locations
        o = locations.process_locations(fixture)
        assert o['1'].level == 0
        assert o['1'].override.bar == 'baz'

    def test_process_locations_levels(self, sample_nested_locations):
        fixture = sample_nested_locations
        o = locations.process_locations(fixture)
        assert o['1'].level == 0
        assert o['bar'].level == 1
        assert o['qux'].level == 2

    def test_generate_location_matrix_cols(self, sample_locations):
        techs = ['demand', 'unmet_demand', 'ccgt', 'csp']
        df = locations.generate_location_matrix(sample_locations, techs)
        wanted_cols = ['_level',
                       '_override.ccgt.constraints.e_cap.max',
                       '_override.demand.constraints.r',
                       '_override.demand.constraints.r_scale_to_peak',
                       '_override.demand.x_map',
                       '_within', 'ccgt', 'csp', 'demand', 'unmet_demand']
        assert sorted(df.columns) == wanted_cols

    def test_generate_location_matrix_index(self, sample_locations):
        techs = ['demand', 'unmet_demand', 'ccgt', 'csp']
        df = locations.generate_location_matrix(sample_locations, techs)
        assert df.index.tolist() == ['test', 'test1']

    def test_generate_location_matrix_values(self, sample_locations):
        techs = ['demand', 'unmet_demand', 'ccgt', 'csp']
        df = locations.generate_location_matrix(sample_locations, techs)
        assert df.at['test', 'demand'] == 1
        assert df.at['test1', 'demand'] == 0
        assert (df.at['test', '_override.demand.constraints.r']
                == 'file=demand.csv')
        assert pd.isnull(df.at['test1', '_override.demand.constraints.r'])

    def test_generate_location_matrix_additional_techs(self, sample_locations):
        techs = ['demand', 'unmet_demand', 'ccgt', 'csp', 'foo']
        df = locations.generate_location_matrix(sample_locations, techs)
        assert sum(df['foo']) == 0

    def test_generate_location_matrix_missing_techs_cols(self, sample_locations):
        techs = ['ccgt']
        df = locations.generate_location_matrix(sample_locations, techs)
        wanted_cols = ['_level',
                       '_override.ccgt.constraints.e_cap.max',
                       '_override.demand.constraints.r',
                       '_override.demand.constraints.r_scale_to_peak',
                       '_override.demand.x_map',
                       '_within', 'ccgt']
        assert sorted(df.columns) == wanted_cols

    def test_generate_location_matrix_within_only_strings(self,
                                                      sample_nested_locations):
        techs = ['foo']
        df = locations.generate_location_matrix(sample_nested_locations, techs)
        for i in df['_within'].tolist():
            assert (i is None or isinstance(i, str))



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

cli.py
~~~~~~

Command-line interface.

"""

import contextlib
import datetime
import logging
import os
import shutil
import sys
import traceback

import click

from . import core
from . import _version
from .parallel import Parallelizer


_debug = click.option(
    '--debug', is_flag=True, default=False,
    help='Print debug information when encountering errors.'
)

_pdb = click.option(
    '--pdb', is_flag=True, default=False,
    help='If used together with --debug, drop into interactive '
         'debugger on encountering errors.'
)

_profile = click.option(
    '--profile', is_flag=True, default=False,
    help='Run through cProfile.'
)

_profile_filename = click.option(
    '--profile_filename', type=str,
    help='Filename to save profile to if enabled --profile.'
)

logging.basicConfig(stream=sys.stderr,
                    format='[%(asctime)s] %(levelname)-8s %(message)s',
                    datefmt=core._time_format)
logger = logging.getLogger()


@contextlib.contextmanager
def format_exceptions(debug=False, pdb=False, profile=False, profile_filename=None, start_time=None):
    if debug:
        logger.setLevel('DEBUG')

    try:
        if profile:
            import cProfile
            profile = cProfile.Profile()
            profile.enable()
        yield
        if profile:
            profile.disable()
            if profile_filename:
                dump_path = os.path.expanduser(profile_filename)
                print('\nSaving cProfile output to: {}'.format(dump_path))
                profile.dump_stats(dump_path)
            else:
                print('\n\n----PROFILE OUTPUT----\n\n')
                profile.print_stats()

    except Exception as e:
        if debug:
            traceback.print_exc()
            if pdb:
                import pdb
                pdb.post_mortem(e.__traceback__)
        else:
            stack = traceback.extract_tb(e.__traceback__)
            # Get last stack trace entry still in Calliope
            last = [i for i in stack if 'calliope' in i[0]][-1]
            if debug:
                err_string = '\nError in {}, {}:{}'.format(last[2], last[0], last[1])
            else:
                err_string = '\nError in {}:'.format(last[2])
            click.secho(err_string, fg='red')
            click.secho(str(e), fg='red')
            if start_time:
                print_end_time(start_time, msg='aborted due to an error')
        sys.exit(1)


def print_end_time(start_time, msg='complete'):
    end_time = datetime.datetime.now()
    secs = round((end_time - start_time).total_seconds(), 1)
    tend = end_time.strftime(core._time_format)
    print('\nCalliope run {}. '
          'Elapsed: {} seconds (time at exit: {})'.format(msg, secs, tend))


def _get_version():
    return 'Version {}'.format(_version.__version__)


@click.group(invoke_without_command=True)
@click.pass_context
@click.option('--version', is_flag=True, default=False,
              help='Display version.')
def cli(ctx, version):
    """Calliope: a multi-scale energy systems (MUSES) modeling framework"""
    if ctx.invoked_subcommand is None and not version:
        print(ctx.get_help())
    if version:
        print(_get_version())


@cli.command(short_help='create model')
@click.argument('path')
@_debug
def new(path, debug):
    """
    Create new model at the given PATH, based on the included example model.
    The directory must not yet exist, and intermediate directories will
    be created automatically.
    """
    if debug:
        print(_get_version())
    # Copies the included example model
    example_model = os.path.join(os.path.dirname(__file__), 'example_model')
    click.echo('Creating new model in: {}'.format(path))
    shutil.copytree(example_model, path)


@cli.command(short_help='directly run single model')
@click.argument('run_config')
@_debug
@_pdb
@_profile
@_profile_filename
def run(run_config, debug, pdb, profile, profile_filename):
    """Execute the given RUN_CONFIG run configuration file."""
    if debug:
        print(_get_version())
    logging.captureWarnings(True)
    start_time = datetime.datetime.now()
    with format_exceptions(debug, pdb, profile, profile_filename, start_time):
        tstart = start_time.strftime(core._time_format)
        print('Calliope run starting at {}\n'.format(tstart))
        model = core.Model(config_run=run_config)
        model.verbose = True  # Enables some print calls inside Model
        model_name = model.config_model.get_key('name', default='None')
        run_name = model.config_run.get_key('name', default='None')
        print('Model name:   {}'.format(model_name))
        print('Run name:     {}'.format(run_name))
        num_techs = (len(model.config_model.techs)
                     - len(core.get_default_techs()))
        msize = '{x} locations, {y} technologies, {t} timesteps'.format(
            x=len(model._sets['x']),
            y=num_techs,
            t=len(model._sets['t']))
        print('Model size:   {}\n'.format(msize))
        if not profile:
            model.config_run.set_key('output.save', True)  # Always save output
        model.run()
        print_end_time(start_time)


@cli.command(short_help='generate parallel runs')
@click.argument('run_config')
@click.argument('path', default='runs')
@click.option('--silent', is_flag=True, default=False,
              help='Be less verbose.')
@_debug
@_pdb
def generate(run_config, path, silent, debug, pdb):
    """
    Generate parallel runs based on the given RUN_CONFIG configuration
    file, saving them in the given PATH, which is a path to a
    directory that must not yet exist (PATH defaults to 'runs'
    if not specified).
    """
    if debug:
        print(_get_version())
    logging.captureWarnings(True)
    with format_exceptions(debug, pdb):
        parallelizer = Parallelizer(target_dir=path, config_run=run_config)
        if not silent and 'name' not in parallelizer.config.parallel:
            click.echo('`' + run_config +
                       '` does not specify a `parallel.name`' +
                       'and was skipped.')
            return
        click.echo('Generating runs from config '
                   '`{}` inside `{}`'.format(run_config, path))
        parallelizer.generate_runs()



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

read.py
~~~~~~~

Functions to read saved model results.

"""

import glob
import logging
import os

import pandas as pd
import xarray as xr

from .utils import AttrDict


REQUIRED_TABLES = ['capacity_factor', 'levelized_cost',
                   'metadata', 'groups', 'shares', 'summary',
                   'time_res']


def _check(path, solution):
    # Superficial check if some key tables are missing
    missing_keys = set(REQUIRED_TABLES) - set(solution.data_vars)
    if len(missing_keys) > 0:
        logging.warning('Solution {} missing tables: '
                        '{}'.format(path, missing_keys))


def read_netcdf(path):
    """Read model solution from NetCDF4 file"""
    with xr.open_dataset(path) as solution:
        solution.load()

    # Deserialize YAML attributes
    for k in ['config_model', 'config_run']:
        solution.attrs[k] = AttrDict.from_yaml_string(solution.attrs[k])

    _check(path, solution)

    return solution


def read_csv(directory):
    solution = AttrDict()
    tables_to_read = glob.glob(directory + '/*.csv')
    if len(tables_to_read) == 0:
        raise IOError('No CSV files found')
    # Only keep basenames without extension
    tables_to_read = [os.path.splitext(os.path.basename(f))[0]
                      for f in tables_to_read]
    arrays = {}
    for f in tables_to_read:
        src = os.path.join(directory, f + '.csv')
        cols = pd.read_csv(src, nrows=1).columns
        series = pd.read_csv(src, index_col=list(range(len(cols) - 1)), parse_dates=True, squeeze=True)
        # Make everything except the last column a MultiIndex
        arrays[f] = xr.DataArray.from_series(series)
    solution = xr.Dataset(arrays)

    # Restore metadata from YAML
    md = AttrDict.from_yaml(os.path.join(directory, 'metadata.yaml'))
    for k in md.keys():
        solution.attrs[k] = md[k]

    _check(directory, solution)

    return solution


def _detect_format(directory):
    """Detects format, falling back to CSV if it can't find NetCDF4"""
    if os.path.exists(os.path.join(directory, 'solution.nc')):
        return 'netcdf'
    else:
        return 'csv'


def read_dir(directory):
    """Combines output files from `directory` and return an AttrDict
    containing them all.

    If a solution is missing or there is an error reading it, an empty
    AttrDict is added to the results in its stead and the error is logged.

    """
    results = AttrDict()
    results.iterations = pd.read_csv(os.path.join(directory, 'iterations.csv'),
                                     index_col=0)
    results.solutions = AttrDict()
    for i in results.iterations.index.tolist():
        iteration_dir = os.path.join(directory, '{:0>4d}'.format(i))
        fmt = _detect_format(iteration_dir)
        logging.debug('Iteration: {}, Format detected: {}'.format(i, fmt))
        try:
            if fmt == 'netcdf':
                sol_path = os.path.join(iteration_dir, 'solution.nc')
                results.solutions[i] = read_netcdf(sol_path)
            else:
                sol_path = iteration_dir
                results.solutions[i] = read_csv(sol_path)
            logging.debug('Read as {}: {}'.format(fmt, sol_path))
        except IOError as err:
            logging.warning('I/O error in `{}` at iteration `{}`'
                            ': {}'.format(iteration_dir, i, err))
            # results.solutions[i] = AttrDict()  # add an empty entry
            continue
    return results

##
# Functionality to post-process parallel runs into aggregated NetCDF files
##


def union_of_indexes(indexes):
    idx = indexes[0]
    for i in range(len(indexes) - 1):
        idx = idx.union(indexes[i + 1])
    return idx


def get_overarching_time_index(datasets):
    all_time_dims = [v.coords['t'].to_index() for k, v in datasets.items()]
    return union_of_indexes(all_time_dims)


def get_longest_time_index_length(datasets):
    all_time_dims = [len(v.coords['t'].to_index()) for k, v in datasets.items()]
    return max(all_time_dims)


def results_to_dataset(results, run_name, reset_time_index=False):
    # Add 'run' dimension
    for k in results.solutions:
        results.solutions[k].coords['run'] = k
        results.solutions[k]['run'] = k

    # Reindex all datasets with reset list of integeres or
    # union of all time indices
    # In the first case, all loaded datasets should have roughly the
    # same length of dimension t (e.g. multiple runs across a single year)
    if reset_time_index:
        max_len_idx = list(range(get_longest_time_index_length(results.solutions)))
        for k in results.solutions:
            results.solutions[k]['t'] = list(range(len(results.solutions[k]['t'])))
            results.solutions[k] = results.solutions[k].reindex(dict(t=max_len_idx))
    else:
        new_idx = get_overarching_time_index(results.solutions)
        for k in results.solutions:
            results.solutions[k] = results.solutions[k].reindex(dict(t=new_idx))

    ds_results = xr.concat(results.solutions.values(), dim='run')

    # Remove metadata
    for k in list(ds_results.attrs.keys()):
        if k != 'calliope_version':
            del ds_results.attrs[k]

    results.iterations.index.name = 'run'
    results.iterations.columns.name = 'cols_iterations'
    ds_results['iterations'] = results.iterations
    ds_results['iterations'] = ds_results.iterations.astype(str)  # Force to str

    # Add run name dimension
    ds_results['run_name'] = xr.DataArray([run_name], coords={'run_name': [run_name]})
    ds_results = ds_results.set_coords('run_name')

    return ds_results


def dir_to_dataset(in_dir, run_name, reset_time_index=False):
    results = read_dir(in_dir)
    return results_to_dataset(results, run_name, reset_time_index)


def convert_run_dir_to_netcdf(in_dir, out_file, reset_time_index=False):
    this_dir = os.path.join(in_dir, 'Output')
    run_name = this_dir.split('/')[-1]
    ds = dir_to_dataset(this_dir, run_name, reset_time_index)

    encoding = {k: {'zlib': True, 'complevel': 4} for k in ds.data_vars}
    ds.to_netcdf(out_file, format='netCDF4', encoding=encoding)
    ds.close()  # Force-close NetCDF file after writing


def convert_subdirs_to_netcdfs(in_dir, out_dir, reset_time_index_for_subdirs=None):
    if reset_time_index_for_subdirs is None:
        reset_time_index_for_subdirs = []

    subdirs = [
        i for i in os.listdir(in_dir)
        if os.path.isdir(os.path.join(in_dir, i))
        and not i.startswith('.')
    ]

    os.makedirs(out_dir, exist_ok=True)

    for s in subdirs:
        if s in reset_time_index_for_subdirs:
            reset_time_index = True
        else:
            reset_time_index = False
        this_path = os.path.join(in_dir, s)
        out_file = os.path.join(out_dir, s + '.nc')
        if os.path.exists(out_file):
            print('File exists, skipping: {}'.format(out_file))
        else:
            print('Processing {}'.format(this_path))
            convert_run_dir_to_netcdf(this_path, out_file, reset_time_index)


def combine_subdir_netcdfs(in_dir, out_file):
    in_files = glob.glob(os.path.join(in_dir, '*.nc'))
    datasets = [xr.open_dataset(i) for i in in_files]

    t_idx = get_overarching_time_index(datasets)
    for i in range(len(datasets)):
        datasets[i] = datasets[i].reindex(dict(t=t_idx))

    ds = xr.concat(datasets, dim='run_name')

    encoding = {k: {'zlib': True, 'complevel': 4} for k in ds.data_vars}
    ds.to_netcdf(out_file, format='netCDF4', encoding=encoding)
    ds.close()  # Force-close NetCDF file after writing
    for d in datasets:
        d.close()



import pytest
import tempfile

from calliope.utils import AttrDict
from . import common
from .common import assert_almost_equal, solver, solver_io


class TestModel:
    @pytest.fixture(scope='module')
    def model(self):
        locations = """
            locations:
                1:
                    techs: ['demand_power']
                    override:
                        demand_power:
                            constraints:
                                r: -50
                sub1,sub2:
                    within: 1
                    techs: ['ccgt']
                    override:
                        ccgt:
                            constraints:
                                e_cap.max: 25
            links:
        """
        config_run = """
            mode: plan
            model: ['{techs}', '{locations}']
            subset_t: ['2005-01-01', '2005-01-02']
        """
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(locations.encode('utf-8'))
            f.read()
            override_dict = AttrDict({
                'solver': solver,
                'solver_io': solver_io,
            })
            model = common.simple_model(config_run=config_run,
                                        config_locations=f.name,
                                        override=override_dict)
        model.run()
        return model

    def test_model_solves(self, model):
         assert str(model.results.solver.termination_condition) == 'optimal'

    def test_model_balanced(self, model):
        sol = model.solution
        assert sol['e'].loc[dict(c='power', y='ccgt')].sum(dim='x').mean() == 50
        assert (sol['e'].loc[dict(c='power', y='ccgt')].sum(dim='x') ==
                -1 * sol['e'].loc[dict(c='power', y='demand_power')].sum(dim='x')).all()

    def test_model_costs(self, model):
        sol = model.solution
        assert_almost_equal(sol['summary'].to_pandas().loc['ccgt', 'levelized_cost_monetary'], 0.1)



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

transmission.py
~~~~~~~~~~~~~~~

Functions to deal with transmission between locations.

"""

from . import locations
from . import utils


def get_remotes(y, x):
    """
    For a given pair of ``y`` (tech) and ``x`` (location), return
    ``(y_remote, x_remote)``, a tuple giving the corresponding indices
    of the remote location a transmission technology is connected to.

    Example: for ``(y, x) = ('hvdc:region_2', 'region_1')``,
    returns ``('hvdc:region_1', 'region_2')``

    """
    y_remote = y.split(':')[0] + ':' + x
    x_remote = y.split(':')[1]
    return (y_remote, x_remote)


def get_transmission_techs(links):
    """Extract a list of all transmission technologies needed for the
    given ``links`` (an AttrDict). Returns an empty list if ``links`` is
    empty.

    """
    if links is None:
        return []
    transmission_y = []
    for i in links:
        for y_base in links[i]:
            for x in locations.explode_locations(i):
                transmission_y.append(y_base + ':' + x)
    # Pass through set to remove duplicates
    return list(set(transmission_y))


def explode_transmission_tree(links, possible_x):
    """Return an AttrDict with configuration for all possible transmission
    technologies defined by ``links``, checking if they have been defined
    to a location within ``possible_x`` (which can be a list or other iterable).

    Returns None if ``links`` empty.

    """
    if links is None:
        return None
    tree = utils.AttrDict()
    for k in links:
        pairs = [locations.explode_locations(k)]
        pairs.append(list(reversed(pairs[0])))
        for x, remote_x in pairs:
            if x not in possible_x:
                raise KeyError('Link to inexistent location: ' + x + '.')
            if x == remote_x:
                raise KeyError('Link must be between different locations: '
                               + x + '.')
            if x not in tree:
                tree[x] = utils.AttrDict()
            for y in links[k]:
                tree[x][y + ':' + remote_x] = links[k][y]
    return tree



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

utils.py
~~~~~~~~

Various utility functions, particularly the AttrDict class (a subclass
of regular dict) used for managing model configuration.

"""

from contextlib import contextmanager
from io import StringIO
import functools
import logging
import os
import importlib
import sys

import numpy as np
import yaml

from . import exceptions


class __Missing(object):
    def __repr__(self):
        return ('MISSING')

    def __nonzero__(self):
        return False


_MISSING = __Missing()


def _yaml_load(src):
    """Load YAML from a file object or path with useful parser errors"""
    if not isinstance(src, str):
        try:
            src_name = src.name
        except AttributeError:
            src_name = '<yaml stringio>'
        # Force-load file streams as that allows the parser to print
        # much more context when it encounters an error
        src = src.read()
    else:
        src_name = '<yaml string>'
    try:
        return yaml.load(src)
    except yaml.YAMLError:
        logging.error('Parser error when reading YAML from {}.'.format(src_name))
        raise


class AttrDict(dict):
    """
    A subclass of ``dict`` with key access by attributes::

        d = AttrDict({'a': 1, 'b': 2})
        d.a == 1  # True

    Includes a range of additional methods to read and write to YAML,
    and to deal with nested keys.

    """
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __init__(self, source_dict=None):
        super(AttrDict, self).__init__()
        if isinstance(source_dict, dict):
            self.init_from_dict(source_dict)

    def copy(self):
        """Override copy method so that it returns an AttrDict"""
        return AttrDict(dict(self).copy())

    def init_from_dict(self, d):
        """
        Initialize a new AttrDict from the given dict. Handles any
        nested dicts by turning them into AttrDicts too::

            d = AttrDict({'a': 1, 'b': {'x': 1, 'y': 2}})
            d.b.x == 1  # True

        """
        for k, v in d.items():
            # First, keys must be strings, not ints
            if isinstance(k, int):
                k = str(k)
            # Now, assign to the key, handling nested AttrDicts properly
            if isinstance(v, dict):
                self.set_key(k, AttrDict(v))
            elif isinstance(v, list):
                self.set_key(k, [i if not isinstance(i, dict) else AttrDict(i)
                                 for i in v])
            else:
                self.set_key(k, v)

    @classmethod
    def from_yaml(cls, f, resolve_imports=True):
        """
        Returns an AttrDict initialized from the given path or
        file object ``f``, which must point to a YAML file.

        If ``resolve_imports`` is True, ``import:`` statements are
        resolved recursively, else they are treated like any other key.

        When resolving import statements, anything defined locally
        overrides definitions in the imported file.

        """
        if isinstance(f, str):
            with open(f, 'r') as src:
                loaded = cls(_yaml_load(src))
        else:
            loaded = cls(_yaml_load(f))
        if resolve_imports and 'import' in loaded:
            for k in loaded['import']:
                imported = cls.from_yaml(relative_path(k, f))
                # loaded is added to imported (i.e. it takes precedence)
                imported.union(loaded)
                loaded = imported
            # 'import' key no longer needed, so we drop it
            loaded.pop('import', None)
        return loaded

    @classmethod
    def from_yaml_string(cls, string):
        """
        Returns an AttrDict initialized from the given string, which
        must be valid YAML.

        """
        return cls(_yaml_load(string))

    def set_key(self, key, value):
        """
        Set the given ``key`` to the given ``value``. Handles nested
        keys, e.g.::

            d = AttrDict()
            d.set_key('foo.bar', 1)
            d.foo.bar == 1  # True

        """
        if '.' in key:
            key, remainder = key.split('.', 1)
            try:
                self[key].set_key(remainder, value)
            except KeyError:
                self[key] = AttrDict()
                self[key].set_key(remainder, value)
            except AttributeError:
                if self[key] is None:  # If the value is None, we replace it
                    self[key] = AttrDict()
                    self[key].set_key(remainder, value)
                # Else there is probably something there, and we don't just
                # want to overwrite so stop and warn the user
                else:
                    raise KeyError('Cannot set nested key on non-dict key.')
        else:
            self[key] = value

    def get_key(self, key, default=_MISSING):
        """
        Looks up the given ``key``. Like set_key(), deals with nested
        keys.

        If default is anything but ``_MISSING``, the given default is
        returned if the key does not exist.

        """
        if '.' in key:
            # Nested key of form "foo.bar"
            key, remainder = key.split('.', 1)
            if default != _MISSING:
                try:
                    value = self[key].get_key(remainder, default)
                except KeyError:
                    # subdict exists, but doesn't contain key
                    return default
                except AttributeError:
                    # key points to non-dict thing, so no get_key attribute
                    return default
            else:
                value = self[key].get_key(remainder)
        else:
            # Single, non-nested key of form "foo"
            if default != _MISSING:
                return self.get(key, default)
            else:
                return self[key]
        return value

    def del_key(self, key):
        """Delete the given key. Properly deals with nested keys."""
        if '.' in key:
            key, remainder = key.split('.', 1)
            try:
                del self[key][remainder]
            except KeyError:
                self[key].del_key(remainder)
        else:
            del self[key]

    def as_dict(self, flat=False):
        """
        Return the AttrDict as a pure dict (with nested dicts if
        necessary).

        """
        if not flat:
            return self.as_dict_nested()
        else:
            return self.as_dict_flat()

    def as_dict_nested(self):
        d = {}
        for k, v in self.items():
            if isinstance(v, AttrDict):
                d[k] = v.as_dict()
            elif isinstance(v, list):
                d[k] = [i if not isinstance(i, AttrDict) else i.as_dict()
                        for i in v]
            else:
                d[k] = v
        return d

    def as_dict_flat(self):
        d = {}
        keys = self.keys_nested()
        for k in keys:
            d[k] = self.get_key(k)
        return d

    def to_yaml(self, path=None, convert_objects=True, **kwargs):
        """
        Saves the AttrDict to the given path as a YAML file.

        If ``path`` is None, returns the YAML string instead.

        Any additional keyword arguments are passed to the YAML writer,
        so can use e.g. ``indent=4`` to override the default of 2.

        ``convert_objects`` (defaults to True) controls whether Numpy
        objects should be converted to regular Python objects, so that
        they are properly displayed in the resulting YAML output.

        """
        if convert_objects:
            result = self.copy()
            for k in result.keys_nested():
                # Convert numpy numbers to regular python ones
                v = result.get_key(k)
                if isinstance(v, np.floating):
                    result.set_key(k, float(v))
                elif isinstance(v, np.integer):
                    result.set_key(k, int(v))
            result = result.as_dict()
        else:
            result = self.as_dict()
        if path is not None:
            with open(path, 'w') as f:
                yaml.dump(result, f, **kwargs)
        else:
            return yaml.dump(result, **kwargs)

    def keys_nested(self, subkeys_as='list'):
        """
        Returns all keys in the AttrDict, sorted, including the keys of
        nested subdicts (which may be either regular dicts or AttrDicts).

        If ``subkeys_as='list'`` (default), then a list of
        all keys is returned, in the form ``['a', 'b.b1', 'b.b2']``.

        If ``subkeys_as='dict'``, a list containing keys and dicts of
        subkeys is returned, in the form ``['a', {'b': ['b1', 'b2']}]``.

        """
        keys = []
        for k, v in sorted(self.items()):
            if isinstance(v, AttrDict) or isinstance(v, dict):
                if subkeys_as == 'list':
                    keys.extend([k + '.' + kk for kk in v.keys_nested()])
                elif subkeys_as == 'dict':
                    keys.append({k: v.keys_nested(subkeys_as=subkeys_as)})
            else:
                keys.append(k)
        return keys

    def union(self, other, allow_override=False, allow_replacement=False):
        """
        Merges the AttrDict in-place with the passed ``other``
        AttrDict. Keys in ``other`` take precedence, and nested keys
        are properly handled.

        If ``allow_override`` is False, a KeyError is raised if
        other tries to redefine an already defined key.

        If ``allow_replacement``, allow "_REPLACE_" key to replace an
        entire sub-dict.

        """
        if allow_replacement:
            WIPE_KEY = '_REPLACE_'
            override_keys = [k for k in other.keys_nested()
                             if WIPE_KEY not in k]
            wipe_keys = [k.split('.' + WIPE_KEY)[0]
                         for k in other.keys_nested()
                         if WIPE_KEY in k]
        else:
            override_keys = other.keys_nested()
            wipe_keys = []
        for k in override_keys:
            if not allow_override and k in self.keys_nested():
                raise KeyError('Key defined twice: {}'.format(k))
            else:
                self.set_key(k, other.get_key(k))
        for k in wipe_keys:
            self.set_key(k, other.get_key(k + '.' + WIPE_KEY))


@contextmanager
def capture_output():
    """
    Capture stdout and stderr output of a wrapped function::

        with capture_output() as out:
            # do things that create stdout or stderr output

    Returns a list with the captured strings: ``[stderr, stdout]``

    """
    old_out, old_err = sys.stdout, sys.stderr
    try:
        out = [StringIO(), StringIO()]
        sys.stdout, sys.stderr = out
        yield out
    finally:
        sys.stdout, sys.stderr = old_out, old_err
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()


# This used to be a custom function, but as of Python 3.2 we can use
# the built-in lru_cache for simplicity
memoize = functools.lru_cache(maxsize=512)


class memoize_instancemethod(object):
    """
    Cache the return value of a method on a per-instance basis
    (as opposed to a per-class basis like functools.lru_cache does)

    Source: http://code.activestate.com/recipes/577452/

    This class is meant to be used as a decorator of methods. The return
    value from a given method invocation will be cached on the instance
    whose method was invoked. All arguments passed to a method decorated
    with memoize must be hashable.

    If a memoized method is invoked directly on its class the result
    will not be cached. Instead the method will be invoked like a
    static method.

    """
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return functools.partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(list(kw.items())))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res


def relative_path(path, base_path_file):
    """
    If ``path`` is not absolute, it is interpreted as relative to the
    path of the given ``base_path_file``.

    """
    # Check if base_path_file is a string because it might be an AttrDict
    if not os.path.isabs(path) and isinstance(base_path_file, str):
        path = os.path.join(os.path.dirname(base_path_file), path)
    return path


def _load_function(source):
    """
    Returns a function from a module, given a source string of the form:

        'module.submodule.subsubmodule.function_name'

    """
    module_string, function_string = source.rsplit('.', 1)
    modules = [i for i in sys.modules.keys() if 'calliope' in i]
    # Check if module already loaded, if so, don't re-import it
    if (module_string in modules):
        module = sys.modules[module_string]
    elif ('calliope.' + module_string) in modules:
        module = sys.modules['calliope.' + module_string]
    # Else load the module
    else:
        try:
            module = importlib.import_module(module_string)
        except ImportError:
            module = importlib.import_module('calliope.' + module_string)
    return getattr(module, function_string)


def plugin_load(name, builtin_module):
    try:  # First try importing as a third-party module
        func = _load_function(name)
    except ValueError:
        # ValueError raised if we got a string without '.',
        # which implies a builtin function,
        # so we attempt to load from the given module
        func_string = builtin_module + '.' + name
        func = _load_function(func_string)
    return func


def option_getter(config_model):
    """Returns a get_option() function using the given config_model and data"""

    def get_option(option, x=None, default=None, ignore_inheritance=False):

        def _get_option(opt, fail=False):
            try:
                result = config_model.get_key('techs.' + opt)
            except KeyError:
                if ignore_inheritance:
                    return _get_option(default, fail)
                # 'ccgt.constraints.s_time' -> 'ccgt', 'constraints.s_time'
                tech, remainder = opt.split('.', 1)
                if ':' in tech:
                    parent = tech.split(':')[0]
                else:
                    # parent = e.g. 'defaults'
                    parent = config_model.get_key('techs.' + tech + '.parent')
                try:
                    result = _get_option(parent + '.' + remainder, fail)
                except KeyError:
                    e = exceptions.OptionNotSetError
                    if fail:
                        raise e('Failed to read option `{}` '
                                'with given default '
                                '`{}`'.format(option, default))
                    elif default:
                        result = _get_option(default, fail=True)
                    elif tech == 'defaults':
                        raise e('Reached top of inheritance chain '
                                'and no default defined for: '
                                '`{}`'.format(option))
                    else:
                        raise e('Can not get parent for `{}` '
                                'and no default defined '
                                '({}).'.format(tech, option))
            return result

        def _get_location_option(key, location):
            # NB1: KeyErrors raised here are always caught in get_option
            # so need no further information or handling
            return config_model.get_key(
                'locations.' + location + '.override.' + key
            )

        if x:
            try:
                result = _get_location_option(option, x)
            # If can't find a location-specific option, fall back to model-wide
            except KeyError:
                result = _get_option(option)
        else:
            result = _get_option(option)
        # Deal with 'inf' settings
        if result == 'inf':
            result = float('inf')
        return result

    return get_option


def cost_getter(option_getter_func, costs_type='costs'):
    def get_cost(cost, y, k, x=None):
        return option_getter_func(
            y + '.' + costs_type + '.' + k + '.' + cost,
            default=y + '.' + costs_type + '.default.' + cost,
            x=x)
    return get_cost


def cost_per_distance_getter(option_getter_func):
    def get_cost_per_distance(cost, y, k, x):
        try:
            cost_option = y + '.costs_per_distance.' + k + '.' + cost
            cost = option_getter_func(cost_option)
            per_distance = option_getter_func(y + '.per_distance')
            distance = option_getter_func(y + '.distance', x=x)
            distance_cost = cost * (distance / per_distance)
        except exceptions.OptionNotSetError:
            distance_cost = 0
        return distance_cost
    return get_cost_per_distance


def depreciation_getter(option_getter_func):
    def get_depreciation_rate(y, k):
        interest = option_getter_func(
            y + '.depreciation.interest.' + k,
            default=y + '.depreciation.interest.default')
        plant_life = option_getter_func(y + '.depreciation.plant_life')
        if interest == 0:
            dep = 1 / plant_life
        else:
            dep = ((interest * (1 + interest) ** plant_life)
                   / (((1 + interest) ** plant_life) - 1))
        return dep
    return get_depreciation_rate


def any_option_getter(model):
    """
    Get any option from the given Model or SolutionModel, including
    ``costs.`` or ``costs_per_distance.`` options

    """
    get_cost = cost_getter(model.get_option)
    get_cost_pd = cost_getter(model.get_option, 'costs_per_distance')

    def get_any_option(option):
        if 'costs.' in option:
            y, rest, k, cost = option.split('.')
            return get_cost(cost, y, k)
        elif 'costs_per_distance.' in option:
            y, rest, k, cost = option.split('.')
            return get_cost_pd(cost, y, k)
        else:
            return model.get_option(option)
    return get_any_option

def vincenty(coord1, coord2):
    """
    Vincenty's inverse method formula to calculate the distance in metres
    between two points on the surface of a spheroid (WGS84).
    modified from https://github.com/maurycyp/vincenty
    """

    a = 6378137  # equitorial radius in meters
    f = 1 / 298.257223563 # flattening from sphere to oblate spheroid
    b = a * (1 - f) # polar radius in meters

    max_iter = 200
    thresh = 1e-12

    # short-circuit coincident points
    if coord1[0] == coord2[0] and coord1[1] == coord2[1]:
        return 0

    U1 = np.arctan((1 - f) * np.tan(np.radians(coord1[0])))
    U2 = np.arctan((1 - f) * np.tan(np.radians(coord2[0])))
    L = np.radians(coord2[1] - coord1[1])
    Lambda = L

    sinU1 = np.sin(U1)
    cosU1 = np.cos(U1)
    sinU2 = np.sin(U2)
    cosU2 = np.cos(U2)

    for iteration in range(max_iter):
        sinLambda = np.sin(Lambda)
        cosLambda = np.cos(Lambda)
        sinSigma = np.sqrt((cosU2 * sinLambda) ** 2 +
                             (cosU1 * sinU2 - sinU1 * cosU2 * cosLambda) ** 2)
        if sinSigma == 0:
            return 0.0  # coincident points
        cosSigma = sinU1 * sinU2 + cosU1 * cosU2 * cosLambda
        sigma = np.arctan2(sinSigma, cosSigma)
        sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
        cosSqAlpha = 1 - sinAlpha ** 2
        try:
            cos2SigmaM = cosSigma - 2 * sinU1 * sinU2 / cosSqAlpha
        except ZeroDivisionError:
            cos2SigmaM = 0
        C = f / 16 * cosSqAlpha * (4 + f * (4 - 3 * cosSqAlpha))
        LambdaPrev = Lambda
        Lambda = L + (1 - C) * f * sinAlpha * (sigma + C * sinSigma *
                                               (cos2SigmaM + C * cosSigma *
                                                (-1 + 2 * cos2SigmaM ** 2)))
        if abs(Lambda - LambdaPrev) < thresh:
            break  # successful convergence
    else:
        return None  # failure to converge

    uSq = cosSqAlpha * (a ** 2 - b ** 2) / (b ** 2)
    A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
    B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))
    deltaSigma = B * sinSigma * (cos2SigmaM + B / 4 * (cosSigma *
                 (-1 + 2 * cos2SigmaM ** 2) - B / 6 * cos2SigmaM *
                 (-3 + 4 * sinSigma ** 2) * (-3 + 4 * cos2SigmaM ** 2)))
    D = b * A * (sigma - deltaSigma)

    return round(D)



import pytest  # pylint: disable=unused-import
import tempfile

from calliope.utils import AttrDict
from . import common
from .common import assert_almost_equal, solver, solver_io, _add_test_path


def create_and_run_model(override, iterative_warmstart=False,
                         demand_file='demand-static_r.csv'):
    locations = """
        locations:
            1:
                techs: ['ccgt', 'demand_power']
                override:
                    ccgt:
                        constraints:
                            e_cap.max: 100
                    demand_power:
                        x_map: '1: demand'
                        constraints:
                            r: file={demand_file}
        links:
    """
    config_run = """
        mode: operate
        model: ['{techs}', '{locations}']
    """
    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    override.set_key('solver_io', solver_io)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(locations.format(demand_file=demand_file).encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override,
                                    path=_add_test_path('common/t_time'))
    model.run(iterative_warmstart)
    return model


class TestModel:
    def test_model_fixed_costs_op(self):
        override = """
            override:
                techs:
                    ccgt:
                        costs:
                            monetary:
                                e_cap: 5
                                om_fuel: 0
            subset_t: ['2005-01-01', '2005-01-0{}']
        """
        model1 = create_and_run_model(override.format(2))
        cost1 = model1.solution['costs'].loc[dict(k='monetary', x='1', y='ccgt')]
        lcoe1 = model1.solution['summary'].loc[dict(techs='ccgt', cols_summary='levelized_cost_monetary')]
        model2 = create_and_run_model(override.format(4))
        cost2 = model2.solution['costs'].loc[dict(k='monetary', x='1', y='ccgt')]
        lcoe2 = model2.solution['summary'].loc[dict(techs='ccgt', cols_summary='levelized_cost_monetary')]
        # LCOE should be the same, as the output is constant throughout
        assert_almost_equal(lcoe1, lcoe2, tolerance=0.0000001)
        # Cost should be double in the second case as it's twice the time
        assert_almost_equal(2 * cost1, cost2, tolerance=0.0000001)

    def test_model_var_costs_op(self):
        override = """
            override:
                techs:
                    ccgt:
                        costs:
                            monetary:
                                e_cap: 0
                                om_fuel: 0.1
            subset_t: ['2005-01-01', '2005-01-0{}']
        """
        model1 = create_and_run_model(override.format(2))
        cost1 = model1.solution['costs'].loc[dict(k='monetary', x='1', y='ccgt')]
        lcoe1 = model1.solution['summary'].loc[dict(techs='ccgt', cols_summary='levelized_cost_monetary')]
        model2 = create_and_run_model(override.format(4))
        cost2 = model2.solution['costs'].loc[dict(k='monetary', x='1', y='ccgt')]
        lcoe2 = model2.solution['summary'].loc[dict(techs='ccgt', cols_summary='levelized_cost_monetary')]
        # LCOE should be the same, as the output is constant throughout
        assert_almost_equal(lcoe1, lcoe2, tolerance=0.0000001)
        # Cost should be double in the second case as it's twice the time
        assert_almost_equal(2 * cost1, cost2, tolerance=0.0000001)

    def test_model_fixed_costs_op_with_varying_demand(self):
        override = """
            override:
                techs:
                    ccgt:
                        costs:
                            monetary:
                                e_cap: 5
                                om_fuel: 0
            subset_t: ['2005-01-01', '2005-01-0{}']
        """
        demand = 'demand-blocky_r.csv'
        model1 = create_and_run_model(override.format(2), demand_file=demand)
        cost1 = model1.solution['costs'].loc[dict(k='monetary', x='1', y='ccgt')]
        lcoe1 = model1.solution['summary'].loc[dict(techs='ccgt', cols_summary='levelized_cost_monetary')]
        model2 = create_and_run_model(override.format(4), demand_file=demand)
        cost2 = model2.solution['costs'].loc[dict(k='monetary', x='1', y='ccgt')]
        lcoe2 = model2.solution['summary'].loc[dict(techs='ccgt', cols_summary='levelized_cost_monetary')]
        # LCOE should be different
        # TODO should check for exact values
        assert lcoe1 > lcoe2  # lcoe1 has lower output, so is higher
        # Cost should be double in the second case as it's twice the time
        assert_almost_equal(2 * cost1, cost2, tolerance=0.0000001)

    def test_model_var_costs_op_with_varying_demand(self):
        override = """
            override:
                techs:
                    ccgt:
                        costs:
                            monetary:
                                e_cap: 0
                                om_fuel: 0.1
            subset_t: ['2005-01-01', '2005-01-0{}']
        """
        demand = 'demand-blocky_r.csv'
        model1 = create_and_run_model(override.format(2), demand_file=demand)
        cost1 = model1.solution['costs'].loc[dict(k='monetary', x='1', y='ccgt')]
        lcoe1 = model1.solution['summary'].loc[dict(techs='ccgt', cols_summary='levelized_cost_monetary')]
        model2 = create_and_run_model(override.format(4), demand_file=demand)
        cost2 = model2.solution['costs'].loc[dict(k='monetary', x='1', y='ccgt')]
        lcoe2 = model2.solution['summary'].loc[dict(techs='ccgt', cols_summary='levelized_cost_monetary')]
        # LCOE should be the same since only fuel costs matter
        assert_almost_equal(lcoe1, lcoe2, tolerance=0.0000001)
        # Cost in second case are more than 2 * first case,
        # because higher output in second case
        assert cost1 == 60
        assert cost2 == 132



import os
import tempfile

import numpy as np
import pytest

import calliope


def verify_solution_integrity(model_solution, solution_from_disk, tempdir):
        # Check whether the two are the same
        np.allclose(model_solution['e_cap'], solution_from_disk['e_cap'])

        # Check that config AttrDict has been deserialized
        assert(solution_from_disk.attrs['config_run'].output.path == tempdir)


class TestSave:
    @pytest.fixture(scope='module')
    def model(self):
        model = calliope.Model()
        model.run()
        return model

    def test_save_netcdf(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            model.config_run.set_key('output.path', tempdir)
            model.save_solution('netcdf')

            # Try reading solution back in
            sol_file = os.path.join(tempdir, 'solution.nc')
            solution_from_disk = calliope.read.read_netcdf(sol_file)
            solution_from_disk.close()  # so that temp dir can be deleted

        verify_solution_integrity(model.solution, solution_from_disk, tempdir)

    def test_save_csv(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            model.config_run.set_key('output.path', tempdir)
            model.save_solution('csv')

            # Try reading solution back in
            solution_from_disk = calliope.read.read_csv(tempdir)

        verify_solution_integrity(model.solution, solution_from_disk, tempdir)



import os
import tempfile

import pytest  # pylint: disable=unused-import
from click.testing import CliRunner

from calliope import cli


class TestCLI:
    def test_new(self):
        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tempdir:
            new_path = os.path.join(tempdir, 'test')
            result = runner.invoke(cli.new, [new_path])
            assert result.exit_code == 0
            # Assert that `run.yaml` in the target dir exists
            assert os.path.isfile(os.path.join(tempdir, 'test', 'run.yaml'))

    def test_run(self):
        runner = CliRunner()
        this_dir = os.path.dirname(__file__)
        run_config = os.path.join(this_dir, '..', 'example_model', 'run.yaml')
        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(cli.run, [run_config])
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, 'Output', 'r.csv'))

    def test_generate(self):
        runner = CliRunner()
        this_dir = os.path.dirname(__file__)
        run_config = os.path.join(this_dir, '..', 'example_model', 'run.yaml')
        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(cli.generate, [run_config])
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, 'runs', 'example-model', 'submit_array.sh'))



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

objective.py
~~~~~~~~~~~~

Objective functions.

"""

import pyomo.core as po  # pylint: disable=import-error


def objective_cost_minimization(model):
    """
    Minimizes total system monetary cost. Used as a default if
    a model does not specify another objective.

    """
    m = model.m

    def obj_rule(m):
        return sum(model.get_option(y + '.weight') *
                   sum(m.cost[y, x, 'monetary'] for x in m.x)
                   for y in m.y)

    m.obj = po.Objective(sense=po.minimize, rule=obj_rule)
    m.obj.domain = po.NonNegativeReals



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

analysis_utils.py
~~~~~~~~~~~~~~~~~

Helper functions for analysis.py.

"""

import itertools

try:
    import matplotlib.pyplot as plt
except ImportError:
    pass  # This is logged in analysis.py
import numpy as np
import pandas as pd

from . import utils


def legend_outside_ax(ax, where='right', artists=None, labels=None, **kwargs):
    """
    Draw a legend outside the figure given by ``ax``.

    """
    box = ax.get_position()

    # Positions depending on `where`
    # 3-tuples: (loc, bbox_to_anchor, positions for ax.set_position())
    POS = {
        'right': ('center left', (1, 0.5),
                  [box.x0, box.y0, box.width * 1.0, box.height]),
        'left': ('upper center', (05., -0.05),
                 [box.x0, box.y0, box.width * 1.0, box.height]),
        'bottom': ('center right', (-0.35, 0.5),
                   [box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9])
    }
    try:
        leg_loc, leg_bbox, new_ax_pos = POS[where]
    except KeyError:
        raise ValueError('Invalid for `where`: {}'.format(where))

    ax.set_position(new_ax_pos)

    # Use custom artists and labels?
    if artists and labels:
        args = [artists, labels]
    else:
        args = []

    legend = ax.legend(*args, loc=leg_loc, bbox_to_anchor=leg_bbox, **kwargs)

    return legend


def stack_plot(df, stack=None, figsize=None, colormap='jet', legend='default',
               ticks='daily', names=None, ax=None, leg_title=None,
               leg_fontsize=None, **kwargs):
    """
    if stack is None, the columns of the passed df are used

    legend can be 'default' or 'right', can set legend title with `leg_title`
    ticks can be 'hourly', 'daily', 'monthly'

    kwargs get passed to ax.stackplot()

    """
    if not stack:
        stack = df.columns
    if not ax:
        if not figsize:
            figsize = (16, 4)
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)

    colors = plt.get_cmap(colormap)(np.linspace(0, 1.0, len(stack))).tolist()
    labels = names if names else stack  # Labels are friendly names, if given
    fills = ax.stackplot(df.index, df[stack].T, colors=colors, labels=labels,
                         **kwargs)

    # A little hack to reverse the ordering of the just-added PolyCollections
    ax.collections = [i for i in ax.collections if i not in fills]
    ax.collections += reversed(fills)

    # Add legend
    if legend:
        if legend == 'default':
            l = ax.legend(title=leg_title)
        elif legend in ['right', 'left', 'bottom']:
            l = legend_outside_ax(ax, where=legend, title=leg_title)
        if leg_title and leg_fontsize:
            plt.setp(l.get_title(), fontsize=leg_fontsize)

    # Format x datetime axis
    # Based on http://stackoverflow.com/a/9627970/397746
    import matplotlib.dates as mdates
    if ticks == 'monthly':
        formatter = mdates.DateFormatter('%b %Y')
        locator = mdates.MonthLocator()
    if ticks == 'daily':
        formatter = mdates.DateFormatter('%d-%m-%Y')
        locator = mdates.DayLocator()
    if ticks == 'hourly':
        formatter = mdates.DateFormatter('%H:%M\n%d-%m-%Y')
        locator = mdates.HourLocator(byhour=[0])
        minor_formatter = mdates.DateFormatter('%H:%M')
        minor_locator = mdates.HourLocator(byhour=list(range(1, 24)))
        plt.gca().xaxis.set_minor_formatter(minor_formatter)
        plt.gca().xaxis.set_minor_locator(minor_locator)
    plt.gca().xaxis.set_major_formatter(formatter)
    plt.gca().xaxis.set_major_locator(locator)
    return ax


def _get_query_string(types):
    formatted_types = ['type == "{}"'.format(t) for t in types]
    query_string = ' | '.join(formatted_types)
    return query_string


def _get_ranges(dates):
    # Modified from http://stackoverflow.com/a/6934267/397746
    while dates:
        end = 1
        timedelta = dates[end] - dates[end - 1]
        try:
            while dates[end] - dates[end - 1] == timedelta:
                end += 1
        except IndexError:
            pass

        yield (dates[0], dates[end - 1])
        dates = dates[end:]


def plot_graph_on_map(config_model, G=None,
                      edge_colors=None, edge_labels=None,
                      figsize=(15, 15), fontsize=9,
                      arrow_style='->',
                      rotate_labels=False,
                      bounds=None,
                      scale_left_distance=0.05,
                      scale_bottom_distance=0.05,
                      ax=None, show_scale=True,
                      map_resolution='i'):
    from mpl_toolkits.basemap import Basemap
    import networkx as nx
    from calliope.lib import nx_pylab

    # Set up basemap
    if not bounds:
        bounds = config_model.metadata.map_boundary
    bounds_width = bounds[2] - bounds[0]  # lon --> width
    bounds_height = bounds[3] - bounds[1]  # lat --> height
    m = Basemap(projection='merc', ellps='WGS84',
                llcrnrlon=bounds[0], llcrnrlat=bounds[1],
                urcrnrlon=bounds[2], urcrnrlat=bounds[3],
                lat_ts=bounds[1] + bounds_width / 2,
                resolution=map_resolution,
                suppress_ticks=True)

    # Node positions
    pos = config_model.metadata.location_coordinates
    pos = {i: m(pos[i][1], pos[i][0]) for i in pos}  # Flip lat, lon to x, y!

    # Create plot
    if not ax:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111, axisbg='w', frame_on=False)
    m.drawmapboundary(fill_color=None, linewidth=0)
    m.drawcoastlines(linewidth=0.2, color='#626262')

    # Draw the graph
    if G:
        # Using nx_pylab to be able to set zorder below the edges
        nx_pylab.draw_networkx_nodes(G, pos, node_color='#CCCCCC',
                                     node_size=300, zorder=0)

        # Using nx_pylab from lib to get arrow_style option
        nx_pylab.draw_networkx_edges(G, pos, width=3,
                                     edge_color=edge_colors,
                                     # This works for edge_use
                                     edge_vmin=0.0, edge_vmax=1.0,
                                     edge_cmap=plt.get_cmap('seismic'),
                                     arrows=True, arrow_style=arrow_style)

        # bbox = dict(color='white', alpha=0.5, edgecolor=None)
        labels = nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels,
                                              rotate=rotate_labels,
                                              font_size=fontsize)

        # Adding node names just above node points
        pos_offset = {i: (pos[i][0], pos[i][1]+20) for i in pos}
        nx.draw_networkx_labels(G, pos_offset, font_size=fontsize)

    # Add a map scale
    if show_scale:
        scale = m.drawmapscale(
            bounds[0] + bounds_width * scale_left_distance,
            bounds[1] + bounds_height * scale_bottom_distance,
            bounds[0], bounds[1],
            100,
            barstyle='simple', labelstyle='simple',
            fillcolor1='w', fillcolor2='#555555',
            fontcolor='#555555', fontsize=fontsize
        )

    return ax, m


def plot_shapefile_polys(shapefile, ax, m, cmap):
    """
    Plot polygons from the given shapefile on a basemap.

    Parameters
    ----------
    shapefile : shapefile object
        Shapefile opened with fiona (``fiona.open(path_to_shapefile)``)
    ax : matplotlib ax object
        Axes to plot on
    m : basemap object
    cmap : matplotlib colormap
        Colormap to use to color the individual polygons in the shapefile

    """
    from descartes import PolygonPatch
    from matplotlib.collections import PatchCollection
    import shapely
    import shapely.ops

    # Break up MultiPolygons into Polygons, keeping track of
    # which shape they belong to
    polygons = []
    names = []
    for i in shapefile:
        # name = i['properties'][name_property]
        name = i['id']
        geom = shapely.geometry.shape(i['geometry'])
        if geom.geom_type == 'MultiPolygon':
            for poly in geom:
                polygons.append(poly)
                names.append(name)
        else:
            polygons.append(geom)
            names.append(name)

    # Create patches with basemap-transformed coordinates
    def TransformedPolygonPatch(poly):
        return PolygonPatch(shapely.ops.transform(m, poly),
                            ec='#555555', lw=.2, alpha=1., zorder=4)

    polygon_patches = [TransformedPolygonPatch(p) for p in polygons]

    # Create name->color_id mapping dict
    colors = {name: idx for idx, name in enumerate(list(set(names)))}
    n = plt.Normalize(0, len(colors))

    pc = PatchCollection(polygon_patches, match_original=True)
    # Now go through all names and set the color appropriately
    pc.set_facecolor(cmap(n([colors[i] for i in names])))
    pc.set_alpha(0.2)
    ax.add_collection(pc)

    return ax


def add_labels(ax, m, label_coordinates, label_positioning, fontsize=13):
    """
    Add labels to a basemap ``m`` on the given ``ax``.

    Parameters
    ----------
    label_coordinates : dict
        Dict of coordinates for each label in the form of
        {label: (lat, lon)}
    label_positioning : dict
        Dict of positions for each label in the form {label: position},
        where position can be 'c' (center), 'r' (right), 'l' (left), or
        'b' (bottom)

    """
    align_x = 25000
    align_y = 35000
    for zone, pos in label_positioning.items():
        lat, lon = label_coordinates[zone]
        x, y = m(lon, lat)
        if pos == 'c':
            # no x/y adjustment for central labels..
            ha = 'center'
        if pos == 'l':
            x -= align_x
            ha = 'right'
        if pos == 'r':
            x += align_x
            ha = 'left'
        if pos == 'b':
            y -= align_y
            ha = 'center'
        ax.add_artist(plt.Text(x, y, zone,
                               fontsize=fontsize,
                               color='black', backgroundcolor='white',
                               horizontalalignment=ha,
                               verticalalignment='bottom',
                               zorder=10))
    return ax


def _get_supply_groups(solution):
    """
    Get individual supply technologies and those groups that define
    group == True, for purposes of calculating diversity of supply

    """
    # idx_1: group is 'True' and '|' in members
    groups = solution.groups.to_pandas()
    grp_1 = groups.query('group == "True" & type == "supply"')
    idx_1 = grp_1[(grp_1.members != grp_1.index)
                  & (grp_1.members.str.contains('\|'))].index.tolist()
    # idx_2: group is 'False' and no '|' in members
    grp_2 = groups.query('group == "False" & type == "supply"')
    idx_2 = grp_2[grp_2.members == grp_2.index].index.tolist()
    # Also drop entries from idx_2 that are already covered by
    # groups in idx_1
    covered = [i.split('|')
               for i in groups.loc[idx_1, 'members'].tolist()]
    covered_flat = [i for i in itertools.chain.from_iterable(covered)]
    idx_2 = [i for i in idx_2 if i not in covered_flat]
    return idx_1 + idx_2


def get_float_formatter_func(precision=None, thousands_sep=False,
                             zero_string='0'):
    """
    Returns a function that gives ``zero_string`` if the float is 0, else
    a string with the given digits of precision.

    """

    def float_formatter(f):
        if isinstance(f, str):
            return f

        if thousands_sep:
            ts = ','
        else:
            ts = ''
        if precision is not None:
            prec = '.' + str(precision) + 'f'
        elif float(f).is_integer():
            prec = '.0f'
        else:
            prec = ''

        fmt = '{:' + ts + prec + '}'

        try:
            if f == 0:
                return zero_string
            elif (precision is not None) or (thousands_sep is not None):
                return fmt.format(f)
            else:
                return str(f)
        except ValueError:
            return f
    return float_formatter


def write_latex_table(df, path, formatters=None, no_textbackslash=False,
                      **kwargs):
    """
    Write a LaTeX table from the given ``df`` (a pandas DataFrame)
    to the given ``path``, optionally using the given ``formatters``
    (see pandas.DataFrame.to_latex documentation).

    Additional **kwargs are passed to pandas.DataFrame.to_latex.

    """
    first_line = '\\begin{tabularx}{0.95\\textwidth}{l' + 'X' * len(df.columns) + '}'
    last_line = '\\end{tabularx}'

    latex_lines = df.to_latex(formatters=formatters, **kwargs).split('\n')
    latex_lines[0] = first_line
    latex_lines[-2] = last_line  # latex_lines[-1] is trailing newline

    # Fix \ chars
    if no_textbackslash:
        latex_lines = [l.replace('\\textbackslash', '\\') for l in latex_lines]

    with open(path, 'w') as f:
        f.write('\n'.join(latex_lines))


def df_transmission_matrix(config_model, tech, constraint='e_cap.max'):
    """
    Returns a transmission matrix in the form of a pandas DataFrame,
    for the given constraint.

    """
    zones = sorted(list(config_model.locations.keys()))

    df = pd.DataFrame(0.0, index=zones, columns=zones)

    for z1 in zones:
        for z2 in zones:
            key = z1 + ',' + z2
            if (key in config_model.links
                    and tech in config_model.links[key]):
                constr_attrdict = config_model.links[key][tech]['constraints']
                val = constr_attrdict.get_key(constraint)
                if val == 'inf':
                    val = pd.np.inf
                df.at[z1, z2] = val
                df.at[z2, z1] = val

    return df


def df_tech_table(model, columns, parent='supply'):
    """
    Returns a pandas DataFrame of technologies from the given model with
    the given parent tech, with  a column for each of the tech
    constraints/options given in ``columns``.

    """
    get_any_option = utils.any_option_getter(model)
    cm = model.config_model
    techs = [k for k in cm.techs
             if model.ischild(k, parent) and 'name' in cm.techs[k]]
    data = []
    for t in techs:
        item = {'name': cm.techs[t].name}
        for c in columns:
            item[c] = get_any_option(t + '.' + c)
        data.append(item)
    return pd.DataFrame.from_records(data)


def flatten(d):
    return {k: d.get_key(k) for k in d.keys_nested()}


def df_locations_table(model, string_in_cols=None, settings_to_get=None):
    """
    Returns a pandas DataFrame of locations from the given model.

    If string_in_cols specified, extracts data from location table.

    If settings_to_get specified, extracts data from model configuration
    directly.

    """
    target = model.config_model.locations
    l = []
    for k in target:
        if string_in_cols:
            item = flatten(target[k])
        if settings_to_get:
            item = {s: model.get_option(s, x=k) for s in settings_to_get}
        item['name'] = k
        l.append(item)
    df = pd.DataFrame.from_records(l)
    if string_in_cols:
        # Always keep the 'name' column
        df = df[[c for c in df.columns if string_in_cols in c] + ['name']]
    return df



import pytest  # pylint: disable=unused-import
import tempfile

from calliope.utils import AttrDict
from . import common
from .common import solver, solver_io, _add_test_path


def create_and_run_model(override, iterative_warmstart=False):
    locations = """
        locations:
            1:
                techs: ['ccgt', 'demand_power']
                override:
                    ccgt:
                        constraints:
                            e_cap.max: 100
                    demand_power:
                        x_map: '1: demand'
                        constraints:
                            r: file=demand-blocky_r.csv
        links:
    """
    config_run = """
        mode: plan
        model: ['{techs}', '{locations}']
    """
    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    override.set_key('solver_io', solver_io)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(locations.encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override,
                                    path=_add_test_path('common/t_time'))
    model.run(iterative_warmstart)
    return model


class TestModel:
    def test_model_subset_t_from_middata(self):
        override = """
            subset_t: ['2005-01-02', '2005-01-03']
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        sol = model.solution
        assert sol['e'].loc[dict(c='power', y='ccgt')].sum(dim=['x', 't']) == 720

    def test_model_time_res_uniform(self):
        override = """
            time: {function: resample, function_options: {'resolution': '1D'}}
        """
        model = create_and_run_model(override)
        assert len(model.data['_time_res']) == 4
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        sol = model.solution
        assert sol['e'].loc[dict(c='power', y='ccgt')].sum(dim=['x', 't']) == 1320

    def test_model_time_res_uniform_subset_t_from_start(self):
        override = """
            time: {function: resample, function_options: {'resolution': '1D'}}
            subset_t: ['2005-01-01', '2005-01-02']
        """
        model = create_and_run_model(override)
        assert len(model.data['_time_res']) == 2
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        sol = model.solution
        assert sol['e'].loc[dict(c='power', y='ccgt')].sum(dim=['x', 't']) == 600

    def test_model_time_res_uniform_subset_t_from_middata(self):
        override = """
            time: {function: resample, function_options: {'resolution': '1D'}}
            subset_t: ['2005-01-02', '2005-01-03']
        """
        model = create_and_run_model(override)
        assert len(model.data['_time_res']) == 2
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        sol = model.solution
        assert sol['e'].loc[dict(c='power', y='ccgt')].sum(dim=['x', 't']) == 720

    def test_model_subset_t_operational(self):
        override = """
            mode: operate
            override:
                opmode:
                    horizon: 24  # Optimization period length (hours)
                    window: 12  # Operation period length (hours)
            subset_t: ['2005-01-02', '2005-01-03']
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        sol = model.solution
        assert sol['e'].loc[dict(c='power', y='ccgt')].sum(dim=['x', 't']) == 720

    def test_model_time_res_uniform_subset_t_operational(self):
        override = """
            mode: operate
            override:
                opmode:
                    horizon: 48  # Optimization period length (hours)
                    window: 24  # Operation period length (hours)
            time: {function: resample, function_options: {'resolution': '12H'}}
            subset_t: ['2005-01-02', '2005-01-03']
        """
        model = create_and_run_model(override)
        assert len(model.data['_time_res']) == 4
        assert str(model.results.solver.termination_condition) == 'optimal'
        # Make sure the result is valid
        sol = model.solution
        assert sol['e'].loc[dict(c='power', y='ccgt')].sum(dim=['x', 't']) == 720



"""
Copyright (C) 2013-2017 Stefan Pfenninger.
Licensed under the Apache 2.0 License (see LICENSE file).

locations.py
~~~~~~~~~~~~

Functions to deal with locations and their configuration.

"""

import pandas as pd

from . import utils


def _generate_location(location, items, techs):
    """
    Returns a dict for a given location. Dict keys for permitted technologies
    are the only ones that don't start with '_'.

    Args:
        location : (str) name of the location
        items : (AttrDict) location settings
        techs : (list) list of available technologies
    """
    # Mandatory basics
    within = items.get('within', None)
    level = items.get('level', None)
    d = {'_location': location, '_level': level,
         '_within': str(within)}
    # Override
    if 'override' in items:
        for k in items.override.keys_nested():
            target_key = '_override.' + k
            try:
                d[target_key] = items.override.get_key(k)
            except KeyError:
                # Allow locations to not define 'within'
                if k == 'within':
                    d[target_key] = None
    # Permitted echnologies
    for y in techs:
        if y in items.techs:
            d[y] = 1
        else:
            d[y] = 0
    return d


def explode_locations(k):
    """Expands the given key ``k``. ``k``s of the form ``'1--3'`` or
    ``'1,2,3'`` are both expanded into the list ``['1', '2', '3']``.

    Can deal with any combination, e.g. ``'1--3,6,9--11,a'`` results in::

        ['1', '2', '3', '6', '9', '10', '11', 'a']

    Always returns a list, even if ``k`` is just a simple key,
    i.e. ``explode_locations('1')`` returns ``['1']``.

    """
    # Ensure sure we don't pass in other things
    assert isinstance(k, str)
    finalkeys = []
    subkeys = k.split(',')
    for sk in subkeys:
        if '--' in sk:
            begin, end = sk.split('--')
            finalkeys += [str(i).strip()
                          for i in range(int(begin), int(end) + 1)]
        else:
            finalkeys += [sk.strip()]
    if finalkeys == [] or finalkeys == ['']:
        raise KeyError('Empty key')
    return finalkeys


def set_location_levels(locations):
    locset = set(locations.keys())

    curr_lvl = 0

    for l in tuple(locset):
        if 'within' not in locations[l] or locations[l].within is None:
            locations[l].level = curr_lvl
            locset.remove(l)

    while len(locset) > 0:
        curr_lvl += 1
        to_subtract = set()
        for l in tuple(locset):
            if locations[l].within not in locset:
                locations[l].level = curr_lvl
                to_subtract.add(l)
        locset = locset - to_subtract

    return locations


def process_locations(d):
    """
    Process locations by taking an AttrDict that may include compact keys
    such as ``1,2,3``, and returning an AttrDict with exactly one key per
    location with all of its settings.

    """
    def _set_loc_key(d, k, value):
        """Set key ``k`` in ``d`` to ``value```."""
        if k in d:
            try:
                d[k].union(value)
            except KeyError as e:
                raise KeyError('Problem at location {}: {}'.format(k, str(e)))
        else:
            d[k] = value
    loc_dict = utils.AttrDict()
    for key in d:
        if ('--' in key) or (',' in key):
            key_locs = explode_locations(key)
            for subkey in key_locs:
                _set_loc_key(loc_dict, subkey, d[key].copy())
        else:
            _set_loc_key(loc_dict, key, d[key].copy())
    locations = set_location_levels(loc_dict)
    return locations


def generate_location_matrix(d, techs):
    """Generate a pandas DataFrame indexed by locations, containing a column
    for each technology in `techs` and 1 if that location is allowed to
    use the technology, else 0.

    The DataFrame also contains _level and _within columns for grouping
    locations into layers and zones.

    """
    rows = []
    for k, v in sorted(d.items()):
        rows.append(_generate_location(k, v, techs))
    df = pd.DataFrame.from_records(rows)
    df.index = df._location
    df = df.drop(['_location'], axis=1)
    return df
